"""
This module contains functions for creating match managers in the Chipiron game framework.
"""

import multiprocessing
from typing import TYPE_CHECKING

import chipiron as ch
import chipiron.games.game as game
import chipiron.players as players
from chipiron.environments.chess_env.board.factory import (
    BoardFactory,
    create_board_factory,
)
from chipiron.games.game.game_manager_factory import GameManagerFactory
from chipiron.games.match.match_args import MatchArgs
from chipiron.games.match.match_manager import MatchManager
from chipiron.games.match.match_results_factory import MatchResultsFactory
from chipiron.players.boardevaluators.factory import create_game_board_evaluator
from chipiron.players.boardevaluators.table_base.factory import (
    AnySyzygyTable,
    create_syzygy,
)
from chipiron.scripts.chipiron_args import ImplementationArgs
from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils import path

from ...environments.chess_env.move_factory import MoveFactory, create_move_factory
from .match_settings_args import MatchSettingsArgs

if TYPE_CHECKING:
    import queue

    from chipiron.players.boardevaluators.board_evaluator import IGameBoardEvaluator
    from chipiron.utils.dataclass import IsDataclass


def create_match_manager(
    args_match: MatchSettingsArgs,
    args_player_one: players.PlayerArgs,
    args_player_two: players.PlayerArgs,
    args_game: game.GameArgs,
    implementation_args: ImplementationArgs,
    universal_behavior: bool = False,
    seed: int | None = None,
    output_folder_path: path | None = None,
    gui: bool = False,
) -> MatchManager:
    """
    Create a match manager for running matches between two players.

    Args:
        implementation_args: (ImplementationArgs) the implementation args
        args_match (MatchSettingsArgs): The match settings arguments.
        args_player_one (players.PlayerArgs): The arguments for player one.
        args_player_two (players.PlayerArgs): The arguments for player two.
        args_game (game.GameArgs): The game arguments.
        seed (int | None, optional): The seed for random number generation. Defaults to None.
        output_folder_path (path | None, optional): The output folder path. Defaults to None.
        gui (bool, optional): Flag indicating whether to enable GUI. Defaults to False.

    Returns:
        MatchManager: The created match manager.
    """
    main_thread_mailbox: queue.Queue[IsDataclass] = multiprocessing.Manager().Queue()

    # Creation of the Syzygy table for perfect play in low pieces cases, needed by the GameManager
    # and can also be used by the players
    syzygy_table: AnySyzygyTable | None = create_syzygy(
        use_rust=implementation_args.use_rust_boards,
    )

    player_one_name: str = args_player_one.name
    player_two_name: str = args_player_two.name

    can_stockfish: bool = args_player_one.name not in [
        "Stockfish"
    ] and args_player_two.name not in ["Stockfish"]
    game_board_evaluator: IGameBoardEvaluator = create_game_board_evaluator(
        gui=gui, can_stockfish=can_stockfish
    )

    board_factory: BoardFactory = create_board_factory(
        use_rust_boards=implementation_args.use_rust_boards,
        use_board_modification=implementation_args.use_board_modification,
        sort_legal_moves=universal_behavior,
    )

    move_factory: MoveFactory = create_move_factory(
        use_rust_boards=implementation_args.use_rust_boards
    )

    game_manager_factory: GameManagerFactory = GameManagerFactory(
        syzygy_table=syzygy_table,
        game_manager_board_evaluator=game_board_evaluator,
        output_folder_path=output_folder_path,
        main_thread_mailbox=main_thread_mailbox,
        board_factory=board_factory,
        implementation_args=implementation_args,
        move_factory=move_factory,
        universal_behavior=universal_behavior,
    )

    match_results_factory: MatchResultsFactory = MatchResultsFactory(
        player_one_name=player_one_name, player_two_name=player_two_name
    )

    game_args_factory: game.GameArgsFactory = game.GameArgsFactory(
        args_match=args_match,
        args_player_one=args_player_one,
        args_player_two=args_player_two,
        seed_=seed,
        args_game=args_game,
    )

    match_manager: MatchManager = MatchManager(
        player_one_id=player_one_name,
        player_two_id=player_two_name,
        game_manager_factory=game_manager_factory,
        game_args_factory=game_args_factory,
        match_results_factory=match_results_factory,
        output_folder_path=output_folder_path,
    )
    return match_manager


def create_match_manager_from_args(
    match_args: MatchArgs,
    base_script_args: BaseScriptArgs,
    implementation_args: ImplementationArgs,
    gui: bool = False,
) -> MatchManager:
    """
    Create a match manager from the given arguments.

    Args:
        implementation_args(ImplementationArgs): The implementation args
        match_args (MatchArgs): The match arguments.
        base_script_args (ScriptArgs) The script arguments.
        gui (bool, optional): Flag indicating whether to enable GUI. Defaults to False.

    Returns:
        MatchManager: The created match manager.
    """

    assert isinstance(match_args.player_one, players.PlayerArgs)
    assert isinstance(match_args.player_two, players.PlayerArgs)

    player_one_args: players.PlayerArgs = match_args.player_one
    player_two_args: players.PlayerArgs = match_args.player_two

    assert isinstance(match_args.match_setting, MatchSettingsArgs)
    assert isinstance(match_args.match_setting.game_args, game.GameArgs)

    assert (
        player_one_args.name != "Command_Line_Human.yaml"
        or not match_args.match_setting.game_args.each_player_has_its_own_thread
    )
    assert (
        player_two_args.name != "Command_Line_Human.yaml"
        or not match_args.match_setting.game_args.each_player_has_its_own_thread
    )

    # taking care of random
    ch.set_seeds(seed=base_script_args.seed)

    match_manager: MatchManager = create_match_manager(
        args_match=match_args.match_setting,
        args_player_one=player_one_args,
        args_player_two=player_two_args,
        output_folder_path=base_script_args.experiment_output_folder,
        universal_behavior=base_script_args.universal_behavior,
        seed=base_script_args.seed,
        args_game=match_args.match_setting.game_args,
        gui=gui,
        implementation_args=implementation_args,
    )

    return match_manager



"""
Module to manage matches.
"""

from .match_factories import create_match_manager
from .match_settings_args import MatchSettingsArgs
from .match_tag import MatchConfigTag

__all__ = ["MatchSettingsArgs", "create_match_manager", "MatchConfigTag"]



"""
Module for the Player class.
"""

from typing import TYPE_CHECKING

from chipiron.environments.chess_env.board import BoardFactory, IBoard
from chipiron.environments.chess_env.board.utils import FenPlusHistory
from chipiron.players.boardevaluators.table_base.factory import AnySyzygyTable
from chipiron.players.move_selector.move_selector import (
    MoveRecommendation,
    MoveSelector,
)
from chipiron.utils import Seed
from chipiron.utils.logger import chipiron_logger

if TYPE_CHECKING:
    from chipiron.environments.chess_env.move import moveUci
    from chipiron.environments.chess_env.move.imove import moveKey

PlayerId = str


class Player:
    """
    Player selects moves on a given board
    """

    #  difference between player and treebuilder includes the fact
    #  that now a player can be a mixture of multiple decision rules
    id: PlayerId
    main_move_selector: MoveSelector
    syzygy: AnySyzygyTable | None
    board_factory: BoardFactory

    def __init__(
        self,
        name: str,
        syzygy: AnySyzygyTable | None,
        main_move_selector: MoveSelector,
        board_factory: BoardFactory,
    ):
        self.id = name
        self.main_move_selector: MoveSelector = main_move_selector
        self.syzygy_player = syzygy
        self.board_factory = board_factory

    def get_id(self) -> PlayerId:
        """
        Returns the ID of the player.

        Returns:
            PlayerId: The player's ID.
        """
        return self.id

    def select_move(
        self, fen_plus_history: FenPlusHistory, seed_int: Seed
    ) -> MoveRecommendation:
        """
        Returns the best move computed by the player.
        The player has the option to ask the syzygy table to play it.

        Args:
            board (BoardChi): The current board state.
            seed_int (seed): The seed for move selection.

        Returns:
            MoveRecommendation: The recommended move.
        """

        board: IBoard = self.board_factory(fen_with_history=fen_plus_history)

        move_recommendation: MoveRecommendation
        # if there is only one possible legal move in the position, do not think, choose it.
        if not board.legal_moves:
            raise ValueError("No legal moves in this position")
        if board.legal_moves.more_than_one_move() or self.id != "Human":
            # if len(list(board.legal_moves))>1 or self.id != 'Human':

            # if the play with syzygy option is on test if the position is in the database to play syzygy
            if self.syzygy_player is not None and self.syzygy_player.fast_in_table(
                board
            ):
                chipiron_logger.info("Playing with Syzygy")
                best_move_key: moveKey = self.syzygy_player.best_move(board)
                best_move_uci: moveUci = board.get_uci_from_move_key(best_move_key)
                move_recommendation = MoveRecommendation(move=best_move_uci)
            else:
                chipiron_logger.info(
                    "Playing with player (not Syzygy) %s\n%s", self.id, board
                )
                move_recommendation = self.main_move_selector.select_move(
                    board=board, move_seed=seed_int
                )
        else:
            move_key: moveKey = list(board.legal_moves)[0]
            move_uci: moveUci = board.get_uci_from_move_key(move_key)
            move_recommendation = MoveRecommendation(move=move_uci)

        return move_recommendation



"""
TreeNodeFactory Protocol
"""

from typing import Protocol

import chipiron.environments.chess_env.board as board_mod
import chipiron.environments.chess_env.board as boards
import chipiron.players.move_selector.treevalue.nodes as node
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode


class TreeNodeFactory[T: ITreeNode](Protocol):
    """
    Interface for Tree Node Factories
    """

    def create(
        self,
        board: boards.IBoard,
        half_move: int,
        count: int,
        parent_node: node.ITreeNode | None,
        move_from_parent: moveKey | None,
        modifications: board_mod.BoardModificationP | None,
    ) -> node.ITreeNode[T]:
        """
        Creates a new TreeNode object.

        Args:
            board (boards.BoardChi): The current state of the chess board.
            half_move (int): The number of half moves made so far in the game.
            count (int): The number of times this position has occurred in the game.
            parent_node (node.ITreeNode | None): The parent node of the new TreeNode.
            move_from_parent (chess.Move | None): The move that led to this position, if any.
            modifications (board_mod.BoardModification | None): The modifications made to the board, if any.

        Returns:
            node.TreeNode: The newly created TreeNode object.
        """
        ...



"""
This module contains the implementation of tree nodes for move selection.

The tree nodes are used in the move selector to represent different moves and their values.

Classes:
- TreeNode: Represents a tree node for move selection.
- ITreeNode: Interface for tree nodes.

"""

from .itree_node import ITreeNode
from .tree_node import TreeNode

__all__ = ["TreeNode", "ITreeNode"]



"""
Defines a factory for creating chess moves, supporting both standard and Rust-based implementations.
"""

from typing import Protocol

import chess
import shakmaty_python_binding

from chipiron.environments.chess_env.board.iboard import IBoard
from chipiron.environments.chess_env.move.imove import IMove

from .board import RustyBoardChi
from .move import moveUci


class MoveFactory(Protocol):
    """Protocol for a move factory."""

    def __call__(self, move_uci: moveUci, board: IBoard | None = None) -> IMove: ...


def create_move_factory(
    use_rust_boards: bool,
) -> MoveFactory:
    """Creates a move factory.

    Args:
        use_rust_boards (bool): Whether to use Rust-based boards.

    Returns:
        MoveFactory: The created move factory.
    """
    move_factory: MoveFactory

    if use_rust_boards:
        # todo can we go back to the not test version without the assert? generics or just typos?
        move_factory = create_rust_move_test_2
    else:
        move_factory = create_move
    return move_factory


def create_rust_move(
    move_uci: moveUci, board: RustyBoardChi | None = None
) -> shakmaty_python_binding.MyMove:
    """Creates a Rust-based move.
    Args:
        move_uci (moveUci): The UCI string of the move.
        board (RustyBoardChi | None, optional): The board on which the move is made. Defaults to None.
    Returns:
        shakmaty_python_binding.MyMove: The created Rust-based move.
    """
    assert board is not None
    return shakmaty_python_binding.MyMove(move_uci, board.chess_)


def create_rust_move_test_2(
    move_uci: moveUci, board: IBoard | None = None
) -> shakmaty_python_binding.MyMove:
    """Creates a Rust-based move for testing purposes.
    Args:
        move_uci (moveUci): The UCI string of the move.
        board (IBoard | None, optional): The board on which the move is made. Defaults to None.
    Returns:
        shakmaty_python_binding.MyMove: The created Rust-based move.
    """
    assert isinstance(board, RustyBoardChi)
    binding_move = shakmaty_python_binding.MyMove(move_uci, board.chess_)
    # rust_move = RustMove(move=binding_move, uci=move_uci)
    # return rust_move
    return binding_move


def create_rust_move_test(move_uci: moveUci, board: IBoard | None = None) -> chess.Move:
    """Creates a Rust-based move for testing purposes.
    Args:
        move_uci (moveUci): The UCI string of the move.
        board (IBoard | None, optional): The board on which the move is made. Defaults to None.
    Returns:
        chess.Move: The created Rust-based move.
    """
    assert board is not None
    return chess.Move.from_uci(move_uci)


def create_move(move_uci: moveUci, board: IBoard | None = None) -> chess.Move:
    """Creates a standard chess move.
    Args:
        move_uci (moveUci): The UCI string of the move.
        board (IBoard | None, optional): The board on which the move is made. Defaults to None.
    Returns:
        chess.Move: The created standard chess move.
    """
    return chess.Move.from_uci(move_uci)



"""
This module contains the game logic and the game arguments.

The `game` module provides the following classes:
- `ObservableGame`: A class that represents an observable game.
- `Game`: A class that represents a game.
- `GameArgs`: A class that represents the game arguments.
- `GameArgsFactory`: A class that provides a factory for creating game arguments.

The module also defines the `__all__` list, which specifies the names that should be imported when using the `from game import *` syntax.

Example usage:
    from game import ObservableGame, Game, GameArgs, GameArgsFactory
"""

from .game import Game, ObservableGame
from .game_args import GameArgs
from .game_args_factory import GameArgsFactory

__all__ = ["GameArgs", "GameArgsFactory", "Game", "ObservableGame"]



"""
init for chipiron
"""

from . import games as game
from . import players as player
from . import utils as tool
from .utils.my_random import set_seeds

__all__ = ["set_seeds", "tool", "game", "player"]



"""
This module contains a GUI script for the chipiron application.
"""

import tkinter as tk
from typing import Any

from chipiron import scripts


def destroy(root: tk.Tk) -> bool:
    """
    Destroy the root window.

    Args:
        root: The root window to be destroyed.

    Returns:
        bool: True if the root window is destroyed successfully.
    """
    root.destroy()
    return True


# TODO switch to pygame


def script_gui() -> tuple[scripts.ScriptType, dict[str, Any]]:
    """
    Run the chipiron GUI script.

    Returns:
        tuple[scripts.ScriptType, dict[str, Any]]: A tuple containing the script type and the GUI arguments.
    """
    root = tk.Tk()
    # place a label on the root window
    root.title("chipiron")
    output: dict[str, Any] = {}

    window_width: int = 800
    window_height: int = 300

    # get the screen dimension
    screen_width: int = root.winfo_screenwidth()
    screen_height: int = root.winfo_screenheight()

    # find the center point
    center_x: int = int(screen_width / 2 - window_width / 2)
    center_y: int = int(screen_height / 2 - window_height / 2)

    # set the position of the window to the center of the screen
    root.geometry(f"{window_width}x{window_height}+{center_x}+{center_y}")
    # root.iconbitmap('download.jpeg')

    message = tk.Label(root, text="What to do?")
    message.grid(column=0, row=0)

    # exit button
    exit_button = tk.Button(root, text="Exit", command=lambda: root.quit())

    message = tk.Label(root, text="Play ")
    message.grid(column=0, row=2)

    # Create the list of options
    color_options_list: list[str] = ["White", "Black"]

    # Variable to keep track of the option
    # selected in OptionMenu
    color_choice_human = tk.StringVar(root)

    # Set the default value of the variable
    color_choice_human.set("White")

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu = tk.OptionMenu(root, color_choice_human, *color_options_list)
    strength_menu.grid(column=1, row=2)

    message = tk.Label(root, text=" against ")
    message.grid(column=2, row=2)

    # Create the list of options
    chipi_algo_options_list: list[str] = ["RecurZipfBase3", "Uniform", "Sequool"]

    # Variable to keep track of the option
    # selected in OptionMenu
    chipi_algo_choice = tk.StringVar(root)

    # Set the default value of the variable
    chipi_algo_choice.set("RecurZipfBase3")

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu = tk.OptionMenu(root, chipi_algo_choice, *chipi_algo_options_list)
    strength_menu.grid(column=1, row=2)

    message = tk.Label(root, text="  strength: ")

    message.grid(column=4, row=2)

    # Create the list of options
    options_list = ["1", "2", "3", "4", "5"]

    # Variable to keep track of the option
    # selected in OptionMenu
    strength_value = tk.StringVar(root)

    # Set the default value of the variable
    strength_value.set("1")

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu = tk.OptionMenu(root, strength_value, *options_list)
    strength_menu.grid(column=5, row=2)

    # play button
    play_against_chipiron_button: tk.Button = tk.Button(
        root,
        text="!Play!",
        command=lambda: [
            play_against_chipiron(
                root,
                strength=strength_value,
                color=color_choice_human,
                chipi_algo=chipi_algo_choice,
            ),
            destroy(root),
        ],
    )

    # watch button
    watch_a_game_button: tk.Button = tk.Button(
        root, text="Watch a game", command=lambda: [watch_a_game(output), destroy(root)]
    )

    # visualize button
    visualize_a_tree_button: tk.Button = tk.Button(
        root,
        text="Visualize a tree",
        command=lambda: [visualize_a_tree(output), destroy(root)],
    )

    play_against_chipiron_button.grid(row=2, column=6)
    watch_a_game_button.grid(row=4, column=0)
    visualize_a_tree_button.grid(row=6, column=0)
    exit_button.grid(row=8, column=0)

    root.mainloop()
    gui_args: dict[str, Any]
    script_type: scripts.ScriptType
    match output["type"]:
        case "play_against_chipiron":
            tree_move_limit = 4 * 10 ** output["strength"]
            gui_args = {
                "config_file_name": "chipiron/scripts/one_match/exp_options.yaml",
                "seed": 0,
                "gui": True,
                "file_name_match_setting": "setting_duda.yaml",
            }
            if output["color_human"] == "White":
                gui_args["file_name_player_one"] = "Human.yaml"
                gui_args["file_name_player_two"] = f"{output['chipi_algo']}.yaml"
                gui_args["player_two"] = {
                    "main_move_selector": {
                        "stopping_criterion": {"tree_move_limit": tree_move_limit}
                    }
                }
            else:
                gui_args["file_name_player_two"] = "Human.yaml"
                gui_args["file_name_player_one"] = f"{output['chipi_algo']}.yaml"
                gui_args["player_one"] = {
                    "main_move_selector": {
                        "stopping_criterion": {"tree_move_limit": tree_move_limit}
                    }
                }
            script_type = scripts.ScriptType.ONE_MATCH
        case "watch_a_game":
            gui_args = {
                "config_file_name": "chipiron/scripts/one_match/exp_options.yaml",
                "seed": 0,
                "gui": True,
                "file_name_player_one": "RecurZipfBase3.yaml",
                "file_name_player_two": "RecurZipfBase4.yaml",
                "file_name_match_setting": "setting_duda.yaml",
            }
            script_type = scripts.ScriptType.ONE_MATCH
        case "tree_visualization":
            gui_args = {
                "config_file_name": "chipiron/scripts/tree_visualization/exp_options.yaml",
            }
            script_type = scripts.ScriptType.TREE_VISUALIZATION
        case other:
            raise ValueError(f"Not a good name: {other}")

    print(f"Gui choices: the script name is {script_type} and the args are {gui_args}")
    return script_type, gui_args


def play_against_chipiron(
    output: tk.Tk, strength: tk.StringVar, color: tk.StringVar, chipi_algo: tk.StringVar
) -> bool:
    """
    Set the output dictionary to indicate playing against chipiron.

    Args:
        output: The output dictionary to be modified.
        strength: The strength of the player.
        color: The color of the human player.
        chipi_algo: The algorithm choice for chipiron.

    Returns:
        bool: True if the output dictionary is modified successfully.
    """
    output["type"] = "play_against_chipiron"
    output["strength"] = int(strength.get())
    output["color_human"] = str(color.get())
    output["chipi_algo"] = str(chipi_algo.get())
    return True


def watch_a_game(output: dict[str, Any]) -> bool:
    """
    Set the output dictionary to indicate watching a game.

    Args:
        output: The output dictionary to be modified.

    Returns:
        bool: True if the output dictionary is modified successfully.
    """
    output["type"] = "watch_a_game"
    return True


def visualize_a_tree(output: dict[str, Any]) -> bool:
    """
    Set the output dictionary to indicate visualizing a tree.

    Args:
        output: The output dictionary to be modified.

    Returns:
        bool: True if the output dictionary is modified successfully.
    """
    output["type"] = "tree_visualization"
    return True



from dataclasses import dataclass
from typing import Any, Literal

import chess
import torch
import torch.nn as nn
from torch.nn import functional as F

from chipiron.players.boardevaluators.neural_networks.NNModelType import NNModelType
from chipiron.utils.chi_nn import ChiNN

number_of_squares = len(chess.SQUARES)
number_pieces_types = len(chess.PIECE_TYPES)
number_colors = len(chess.COLORS)
number_occupancy_types = (
    number_pieces_types * number_colors + 1
)  # could be empty (+1) or one of the pieces in black or white
len_square_tensor = number_of_squares * number_occupancy_types
len_all_possible_tensor_input = (
    len_square_tensor + 1
)  # +1 for the vector that will bear the output embedding
number_parallel_tracks = number_of_squares + 1


@dataclass()
class TransformerArgs:
    """Transformer model hyperparameters."""

    number_occupancy_types: int = number_occupancy_types
    len_square_tensor: int = len_square_tensor
    number_pieces_types: int = number_pieces_types

    type: Literal[NNModelType.TRANSFORMER] = NNModelType.TRANSFORMER
    n_embd: int = 27
    n_head: int = 3
    n_layer: int = 2
    dropout_ratio: float = 0.0

    def __str__(self) -> str:
        """
        Returns a string representation of the TransformerArgs instance.
        """
        return (
            f"TransformerArgs(n_embd={self.n_embd}, "
            f"n_head={self.n_head}, "
            f"n_layer={self.n_layer}, "
            f"dropout_ratio={self.dropout_ratio})"
        )

    def filename(self) -> str:
        """
        Generates a filename based on the TransformerArgs instance.
        """
        return (
            f"transformer_{self.n_embd}embd_"
            f"{self.n_head}head_{self.n_layer}layer_"
            f"{self.dropout_ratio:.2f}dropout"
        )


class Head(nn.Module):
    """one head of self-attention"""

    def __init__(self, n_embd: int, head_size: int, dropout_ratio: float):
        super().__init__()
        self.key = nn.Linear(n_embd, head_size, bias=True)
        self.query = nn.Linear(n_embd, head_size, bias=True)
        self.value = nn.Linear(n_embd, head_size, bias=True)

        self.dropout = nn.Dropout(dropout_ratio)
        # todo investigate gap

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Applies self-attention to the input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (batch, time-step, channels).

        Returns:
            torch.Tensor: Output tensor of shape (batch, time-step, head size).
        """
        # input of size (batch, time-step, channels)
        # output of size (batch, time-step, head size)
        # B, T, C = x.shape
        k = self.key(x)  # (B,T,hs)
        q = self.query(x)  # (B,T,hs)
        # compute attention scores ("affinities")
        wei = (
            q @ k.transpose(-2, -1) * k.shape[-1] ** -0.5
        )  # (B, T, hs) @ (B, hs, T) -> (B, T, T)
        wei = F.softmax(wei, dim=-1)  # (B, T, T)
        wei = self.dropout(wei)
        # perform the weighted aggregation of the values
        v = self.value(x)  # (B,T,hs)
        out: torch.Tensor = wei @ v  # (B, T, T) @ (B, T, hs) -> (B, T, hs)
        return out


class MultiHeadAttention(nn.Module):
    """multiple heads of self-attention in parallel"""

    def __init__(
        self, num_heads: int, head_size: int, dropout_ratio: float, n_embd: int
    ) -> None:
        super().__init__()
        self.heads = nn.ModuleList(
            [
                Head(head_size=head_size, n_embd=n_embd, dropout_ratio=dropout_ratio)
                for _ in range(num_heads)
            ]
        )  # (B, T, hs) * num_heads
        self.proj = nn.Linear(
            head_size * num_heads, n_embd
        )  # (B, T, hs) * num_heads -> (B, T, n_embd)
        self.dropout = nn.Dropout(dropout_ratio)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out: torch.Tensor = torch.empty(1)  # to make mypy and jit happy
        out = torch.cat([h(x) for h in self.heads], dim=-1)
        out = self.dropout(self.proj(out))
        return out


class FeedFoward(nn.Module):
    """a simple linear layer followed by a non-linearity"""

    def __init__(self, n_embd: int, dropout_ratio: float) -> None:
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(n_embd, 4 * n_embd),
            nn.GELU(),
            nn.Linear(4 * n_embd, n_embd),
            nn.Dropout(dropout_ratio),
        )

    # self.lin =   nn.Linear(n_embd, 4 * n_embd)
    # self.re1 =     nn.GELU()
    # self.lin2 =    nn.Linear(4 * n_embd, n_embd)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # xx=self.re1(self.lin(x))
        # a=self.lin2(xx)
        a: torch.Tensor = self.net(x)

        # print('x',x, a, 'yy',self.net[0].weight, 'rr', self.net[0].weight.grad)
        return a


class Block(nn.Module):
    """Transformer block: communication followed by computation"""

    def __init__(self, n_embd: int, n_head: int, dropout_ratio: float) -> None:
        # n_embd: embedding dimension, n_head: the number of heads we'd like
        super().__init__()
        head_size = n_embd // n_head
        self.sa = MultiHeadAttention(
            n_head, head_size, dropout_ratio=dropout_ratio, n_embd=n_embd
        )  # (B, T, n_embd) -> (B, T, n_embd)
        self.ffwd = FeedFoward(
            n_embd, dropout_ratio=dropout_ratio
        )  # (B, T, n_embd) -> (B, T, n_embd)
        self.ln1 = nn.LayerNorm(n_embd)
        self.ln2 = nn.LayerNorm(n_embd)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # print('zzyyy',x,self.ln1(x))

        y = self.sa(self.ln1(x))
        x = x + y
        # print('yyy',y,x,self.ln2(x))
        x = x + self.ffwd(self.ln2(x))
        return x


class TransformerOne(ChiNN):
    def __init__(self, args: TransformerArgs) -> None:
        super(TransformerOne, self).__init__()

        self.board_embedding_table = nn.Parameter(
            torch.randn(len_all_possible_tensor_input, args.n_embd)
        )

        self.blocks = nn.Sequential(
            *[
                Block(args.n_embd, n_head=args.n_head, dropout_ratio=args.dropout_ratio)
                for _ in range(args.n_layer)
            ]
        )
        self.ln_f = nn.LayerNorm(args.n_embd)  # final layer norm
        self.lm_head = nn.Linear(number_parallel_tracks * args.n_embd, 1)

        self.tan_h = nn.Tanh()

        # better init, not covered in the original GPT video, but important, will cover in followup video
        self.apply(self._init_weights)

    def _init_weights(self, module: Any) -> None:
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            torch.nn.init.zeros_(module.bias)

    def forward(self, indices: list[list[int]]) -> torch.Tensor:
        # idx and targets are both (B,T) tensor of integers
        y = self.board_embedding_table[
            indices, :
        ]  # (B,len_all_possible_tensor_input,n_embd)
        z = self.blocks(y)  # (B,T,C)
        # w = self.ln_f(z)  # (B,T,C)
        x: torch.Tensor = self.lm_head(z.flatten(start_dim=1))  # (B,T,vocab_size)
        # print("z",z,"oo", z.sum())

        x = self.tan_h(x)

        return x

    def compute_representation(
        self, node: Any, parent_node: Any, board_modifications: Any
    ) -> None:
        """
        Compute the input representation for the given node.

        Args:
            node (Any): Current node.
            parent_node (Any): Parent node.
            board_modifications (Any): Board modifications.
        """
        ...
        raise Exception(f"to be recoded in {__name__}")

    def get_nn_input(self, node: Any) -> None:
        """
        Get the input tensor for the given node.

        Args:
            node (Any): Current node.

        Returns:
            None
        """
        raise Exception(f"to be recoded in {__name__}")

    def print_param(self) -> None:
        """
        Print the parameters of the model.
        """
        print(
            "print param not implmeented but is it necessary given the summary thingy in pythorch?"
        )



"""This module contains the classes for match results."""

from dataclasses import dataclass
from typing import Protocol

from chipiron.environments.chess_env.move import moveUci
from chipiron.games.game.final_game_result import FinalGameResult


@dataclass
class SimpleResults:
    """
    Represents the simple results of a match.
    """

    player_one_wins: int
    player_two_wins: int
    draws: int


class IMatchResults(Protocol):
    """
    Interface for match results.
    """

    def add_result_one_game(
        self, white_player_name_id: str, game_result: FinalGameResult
    ) -> None:
        """
        Adds the result of one game to the match results.

        Args:
            white_player_name_id (str): The ID of the white player.
            game_result (FinalGameResult): The result of the game.
        """

    def get_simple_result(self) -> SimpleResults:
        """
        Returns the simple results of the match.

        Returns:
            SimpleResults: The simple results of the match.
        """
        ...

    def __str__(self) -> str:
        """
        Returns a string representation of the match results.

        Returns:
            str: A string representation of the match results.
        """
        ...

    def finish(self) -> None:
        """
        Finishes the match and marks it as finished.
        """


@dataclass
class MatchResults:
    """
    Represents the results of a match between two players.
    """

    player_one_name_id: str
    player_two_name_id: str
    number_of_games: int = 0
    player_one_is_white_white_wins: int = 0
    player_one_is_white_black_wins: int = 0
    player_one_is_white_draws: int = 0
    player_two_is_white_white_wins: int = 0
    player_two_is_white_black_wins: int = 0
    player_two_is_white_draws: int = 0
    match_finished: bool = False

    def get_player_one_wins(self) -> int:
        """
        Returns the number of wins for player one.

        Returns:
            int: The number of wins for player one.
        """
        return self.player_one_is_white_white_wins + self.player_two_is_white_black_wins

    def get_player_two_wins(self) -> int:
        """
        Returns the number of wins for player two.

        Returns:
            int: The number of wins for player two.
        """
        return self.player_one_is_white_black_wins + self.player_two_is_white_white_wins

    def get_draws(self) -> int:
        """
        Returns the number of draws.

        Returns:
            int: The number of draws.
        """
        return self.player_one_is_white_draws + self.player_two_is_white_draws

    def get_simple_result(self) -> SimpleResults:
        """
        Returns the simple results of the match.

        Returns:
            SimpleResults: The simple results of the match.
        """
        simple_result: SimpleResults = SimpleResults(
            player_one_wins=self.get_player_one_wins(),
            player_two_wins=self.get_player_two_wins(),
            draws=self.get_draws(),
        )
        return simple_result

    def add_result_one_game(
        self, white_player_name_id: str, game_result: FinalGameResult
    ) -> None:
        """
        Adds the result of one game to the match results.

        Args:
            white_player_name_id (str): The ID of the white player.
            game_result (FinalGameResult): The result of the game.
        """
        self.number_of_games += 1
        if white_player_name_id == self.player_one_name_id:
            if game_result == FinalGameResult.WIN_FOR_WHITE:
                self.player_one_is_white_white_wins += 1
            elif game_result == FinalGameResult.WIN_FOR_BLACK:
                self.player_one_is_white_black_wins += 1
            elif game_result == FinalGameResult.DRAW:
                self.player_one_is_white_draws += 1
            else:
                pass
                # raise Exception('#')
        elif white_player_name_id == self.player_two_name_id:
            if game_result == FinalGameResult.WIN_FOR_WHITE:
                self.player_two_is_white_white_wins += 1
            elif game_result == FinalGameResult.WIN_FOR_BLACK:
                self.player_two_is_white_black_wins += 1
            elif game_result == FinalGameResult.DRAW:
                self.player_two_is_white_draws += 1
            else:
                raise Exception("!")
        else:
            raise Exception("?")

    def finish(self) -> None:
        """
        Finishes the match and marks it as finished.
        """
        self.match_finished = True

    def __str__(self) -> str:
        """
        Returns a string representation of the match results.

        Returns:
            str: A string representation of the match results.
        """
        str_ = (
            "Main result: "
            + self.player_one_name_id
            + " wins "
            + str(self.get_player_one_wins())
            + " "
        )
        str_ += self.player_two_name_id + " wins " + str(self.get_player_two_wins())
        str_ += " draws " + str(self.get_draws()) + "\n"

        str_ += self.player_one_name_id + " with white: "
        str_ += "Wins " + str(self.player_one_is_white_white_wins)
        str_ += ", Losses " + str(self.player_one_is_white_black_wins)
        str_ += ", Draws " + str(self.player_one_is_white_draws)
        str_ += "\n           with black: "
        str_ += "Wins " + str(self.player_two_is_white_black_wins)
        str_ += ", Losses " + str(self.player_two_is_white_white_wins)
        str_ += ", Draws " + str(self.player_two_is_white_draws) + "\n"

        str_ += self.player_two_name_id + " with white: "
        str_ += "Wins " + str(self.player_two_is_white_white_wins)
        str_ += ", Losses " + str(self.player_two_is_white_black_wins)
        str_ += ", Draws " + str(self.player_two_is_white_draws)
        str_ += "\n           with black: "
        str_ += "Wins " + str(self.player_one_is_white_black_wins)
        str_ += ", Losses " + str(self.player_one_is_white_white_wins)
        str_ += ", Draws " + str(self.player_one_is_white_draws)
        return str_


@dataclass
class MatchReport:
    """
    Represents a match report containing the match results and move history.
    """

    match_results: MatchResults
    match_move_history: dict[int, list[moveUci]]



"""
This module defines the Descendants and RangedDescendants classes.

Descendants:
- Represents a collection of descendants of a tree node at different half moves.
- Provides methods to add, remove, and access descendants.
- Keeps track of the number of descendants and the number of descendants at each half move.

RangedDescendants:
- Inherits from Descendants and adds the ability to track a range of half moves.
- Provides methods to check if a half move is within the current range or acceptable range.
- Allows adding and removing descendants within the range.
- Provides a method to get the range of half moves.

Note: The Descendants and RangedDescendants classes are used in the chipiron project for move selection in a game.
"""

import typing
from typing import Iterator

from sortedcollections import ValueSortedDict

import chipiron.environments.chess_env.board as boards
from chipiron.environments import HalfMove
from chipiron.players.move_selector.treevalue.nodes import ITreeNode
from chipiron.players.move_selector.treevalue.nodes.tree_traversal import (
    get_descendants,
)


class Descendants:
    """
    Represents a collection of descendants for a specific half move in a tree.

    Attributes:
        descendants_at_half_move (dict[HalfMove, dict[str, ITreeNode]]): A dictionary that maps a half move to a dictionary of descendants.
        number_of_descendants (int): The total number of descendants in the collection.
        number_of_descendants_at_half_move (dict[HalfMove, int]): A dictionary that maps a half move to the number of descendants at that half move.
        min_half_move (int | None): The minimum half move in the collection, or None if the collection is empty.
        max_half_move (int | None): The maximum half move in the collection, or None if the collection is empty.
    """

    descendants_at_half_move: dict[HalfMove, dict[boards.boardKey, ITreeNode]]
    number_of_descendants: int
    number_of_descendants_at_half_move: dict[HalfMove, int]
    min_half_move: int | None
    max_half_move: int | None

    def __init__(self) -> None:
        """
        Initializes a Descendants object.

        This method initializes the Descendants object by setting up the necessary attributes.

        Attributes:
        - descendants_at_half_move (dict): A dictionary to store the descendants at each half move.
        - number_of_descendants (int): The total number of descendants.
        - number_of_descendants_at_half_move (dict): A dictionary to store the number of descendants at each half move.
        - min_half_move (int or None): The minimum half move.
        - max_half_move (int or None): The maximum half move.
        """

        self.descendants_at_half_move = {}
        self.number_of_descendants = 0
        self.number_of_descendants_at_half_move = {}
        self.min_half_move = None
        self.max_half_move = None

    def iter_on_all_nodes(
        self,
    ) -> Iterator[tuple[HalfMove, boards.boardKey, ITreeNode]]:
        return (
            (hm, board_key, node)
            for hm, nodes_at_hm in self.descendants_at_half_move.items()
            for board_key, node in nodes_at_hm.items()
        )

    def keys(self) -> typing.KeysView[HalfMove]:
        """
        Returns a view of the keys in the descendants_at_half_move dictionary.

        Returns:
            typing.KeysView[HalfMove]: A view of the keys in the descendants_at_half_move dictionary.
        """
        return self.descendants_at_half_move.keys()

    def __setitem__(
        self, half_move: HalfMove, value: dict[boards.boardKey, ITreeNode]
    ) -> None:
        """
        Sets the descendants at a specific half move.

        Args:
            half_move (HalfMove): The half move at which to set the descendants.
            value (dict[str, ITreeNode]): The descendants to set.

        Returns:
            None
        """
        self.descendants_at_half_move[half_move] = value

    def __getitem__(self, half_move: HalfMove) -> dict[boards.boardKey, ITreeNode]:
        """
        Retrieve the descendants at a specific half move.

        Args:
            half_move (HalfMove): The half move to retrieve the descendants for.

        Returns:
            dict[str, ITreeNode]: A dictionary of descendants at the specified half move.
        """
        return self.descendants_at_half_move[half_move]

    def __iter__(self) -> typing.Iterator[HalfMove]:
        """
        Returns an iterator over the descendants at each half move.

        Returns:
            An iterator over the descendants at each half move.
        """
        return iter(self.descendants_at_half_move)

    def get_count(self) -> int:
        """
        Returns the number of descendants for the current node.

        Returns:
            int: The number of descendants.
        """
        return self.number_of_descendants

    def contains_node(self, node: ITreeNode) -> bool:
        """
        Checks if the descendants contain a specific node.

        Args:
            node (ITreeNode): The node to check for.

        Returns:
            bool: True if the descendants contain the node, False otherwise.
        """
        if (
            node.half_move in self.descendants_at_half_move
            and node.fast_rep in self[node.half_move]
        ):
            return True
        else:
            return False

    def remove_descendant(self, node: ITreeNode) -> None:
        """
        Removes a descendant node from the tree.

        Args:
            node (ITreeNode): The node to be removed.

        Returns:
            None
        """
        half_move = node.half_move
        fen = node.fast_rep

        self.number_of_descendants -= 1
        self[half_move].pop(fen)
        self.number_of_descendants_at_half_move[half_move] -= 1
        if self.number_of_descendants_at_half_move[half_move] == 0:
            self.number_of_descendants_at_half_move.pop(half_move)
            self.descendants_at_half_move.pop(half_move)

    def empty(self) -> bool:
        """
        Check if the descendants are empty.

        Returns:
            bool: True if the number of descendants is 0, False otherwise.
        """
        return self.number_of_descendants == 0

    def add_descendant(self, node: ITreeNode) -> None:
        """
        Adds a descendant node to the tree.

        Args:
            node (ITreeNode): The descendant node to be added.

        Returns:
            None
        """
        half_move: HalfMove = node.half_move
        board_key: boards.boardKey = node.fast_rep

        if half_move in self.descendants_at_half_move:
            assert board_key not in self.descendants_at_half_move[half_move]
            self.descendants_at_half_move[half_move][board_key] = node
            self.number_of_descendants_at_half_move[half_move] += 1
        else:
            self.descendants_at_half_move[half_move] = {board_key: node}
            self.number_of_descendants_at_half_move[half_move] = 1
        self.number_of_descendants += 1

    def __len__(self) -> int:
        """
        Returns the number of descendants at the current half move.

        :return: The number of descendants at the current half move.
        :rtype: int
        """
        return len(self.descendants_at_half_move)

    def print_info(self) -> None:
        """
        Prints information about the descendants.

        This method prints the number of descendants and their corresponding half moves.
        It also prints the ID and fast representation of each descendant.

        Returns:
            None
        """
        print("---here are the ", self.get_count(), " descendants.")
        for half_move in self:
            print(
                "half_move: ",
                half_move,
                "| (",
                self.number_of_descendants_at_half_move[half_move],
                "descendants)",
            )  # ,                  end='| ')
            for descendant in self[half_move].values():
                print(descendant.id, descendant.fast_rep, end=" ")
            print("")

    def print_stats(self) -> None:
        """
        Prints the statistics of the descendants.

        This method prints the number of descendants at each half move.

        Returns:
            None
        """
        print("---here are the ", self.get_count(), " descendants")
        for half_move in self:
            print(
                "half_move: ",
                half_move,
                "| (",
                self.number_of_descendants_at_half_move[half_move],
                "descendants)",
            )

    def test(self) -> None:
        """
        This method performs a series of assertions to validate the descendants data structure.
        It checks if the number of descendants at each half move matches the number of descendants stored.
        It also checks if the sum of the lengths of all descendants at each half move matches the total number of descendants.
        """
        assert set(self.descendants_at_half_move.keys()) == set(
            self.number_of_descendants_at_half_move
        )
        sum_ = 0
        for half_move in self:
            sum_ += len(self[half_move])
        assert self.number_of_descendants == sum_

        for half_move in self:
            assert self.number_of_descendants_at_half_move[half_move] == len(
                self[half_move]
            )

    def test_2(self, root_node: ITreeNode) -> None:
        """
        Test the descendants of a given root node.

        Args:
            root_node (ITreeNode): The root node to test.

        Returns:
            None
        """
        all_descendants = get_descendants(root_node)

        # self.print_info()
        for d in all_descendants:
            if d.half_move not in self.descendants_at_half_move:
                assert d.half_move in self.descendants_at_half_move
            if d.fast_rep not in self.descendants_at_half_move[d.half_move]:
                assert d.fast_rep in self.descendants_at_half_move[d.half_move]

        for half_move in self.descendants_at_half_move:
            for d in self[half_move].values():
                assert d in all_descendants


class RangedDescendants(Descendants):
    """
    Represents a collection of descendants with a range of half moves.

    Attributes:
        min_half_move (int | None): The minimum half move in the range.
        max_half_move (int | None): The maximum half move in the range.
    """

    min_half_move: int | None
    max_half_move: int | None

    def __init__(self) -> None:
        """
        Initializes a Descendants object.
        """
        super().__init__()
        self.min_half_move = None
        self.max_half_move = None

    def __str__(self) -> str:
        """
        Returns a string representation of the Descendants object.

        The string includes information about each half move and its descendants.

        Returns:
            str: A string representation of the Descendants object.
        """
        string: str = ""
        for half_move in self:
            string += f"half_move: {half_move} | ({self.number_of_descendants_at_half_move[half_move]} descendants)\n"
            for descendant in self[half_move].values():
                string += f"{descendant.id} "
            string += "\n"
        return string

    def is_new_generation(self, half_move: HalfMove) -> bool:
        """
        Checks if the given half move is a new generation.

        Args:
            half_move (HalfMove): The half move to check.

        Returns:
            bool: True if the half move is a new generation, False otherwise.
        """
        if self.min_half_move is not None:
            assert self.max_half_move is not None
            return half_move == self.max_half_move + 1
        else:
            return True

    def is_in_the_current_range(self, half_move: int) -> bool:
        """
        Checks if the given half_move is within the current range.

        Args:
            half_move (int): The half_move to check.

        Returns:
            bool: True if the half_move is within the range, False otherwise.
        """

        if self.min_half_move is not None and self.max_half_move is not None:
            return self.max_half_move >= half_move >= self.min_half_move
        else:
            return False

    def is_in_the_acceptable_range(self, half_move: int) -> bool:
        """
        Checks if the given half_move is within the acceptable range.

        Args:
            half_move (int): The half_move to check.

        Returns:
            bool: True if the half_move is within the acceptable range, False otherwise.
        """
        if self.min_half_move is not None and self.max_half_move is not None:
            return self.max_half_move + 1 >= half_move >= self.min_half_move
        else:
            return True

    def add_descendant(self, node: ITreeNode) -> None:
        """
        Adds a descendant node to the tree.

        Args:
            node (ITreeNode): The descendant node to be added.

        Returns:
            None
        """
        half_move: int = node.half_move
        board_key: boards.boardKey = node.fast_rep

        assert self.is_in_the_acceptable_range(half_move)
        if self.is_in_the_current_range(half_move):
            if half_move in self.descendants_at_half_move:
                assert board_key not in self.descendants_at_half_move[half_move]
            self.descendants_at_half_move[half_move][board_key] = node
            self.number_of_descendants_at_half_move[half_move] += 1
        else:  # half_move == len(self.descendants_at_half_move)
            assert self.is_new_generation(half_move)
            self.descendants_at_half_move[half_move] = {board_key: node}
            self.number_of_descendants_at_half_move[half_move] = 1
            if self.max_half_move is not None:
                self.max_half_move += 1
            else:
                self.min_half_move = half_move
                self.max_half_move = half_move
        self.number_of_descendants += 1

    def remove_descendant(self, node: ITreeNode) -> None:
        """
        Removes a descendant node from the tree.

        Args:
            node (ITreeNode): The node to be removed.

        Returns:
            None
        """
        half_move: int = node.half_move
        fen = node.fast_rep

        self.number_of_descendants -= 1
        self[half_move].pop(fen)
        self.number_of_descendants_at_half_move[half_move] -= 1
        if self.number_of_descendants_at_half_move[half_move] == 0:
            self.number_of_descendants_at_half_move.pop(half_move)
            self.descendants_at_half_move.pop(half_move)
            if half_move == self.max_half_move:
                assert self.max_half_move is not None
                self.max_half_move -= 1
            if half_move == self.min_half_move:
                assert self.min_half_move is not None
                self.min_half_move += 1
            assert self.max_half_move is not None
            assert self.min_half_move is not None
            if self.max_half_move < self.min_half_move:
                self.max_half_move = None
                self.min_half_move = None
                assert self.number_of_descendants == 0

    def range(self) -> range:
        """
        Returns a range object representing the half moves range.

        The range starts from the minimum half move and ends at the maximum half move.

        Returns:
            range: A range object representing the half moves range.
        """

        assert self.max_half_move is not None
        assert self.min_half_move is not None
        return range(self.min_half_move, self.max_half_move + 1)

    #  def update(
    #          self,
    #          new_descendants: typing.Self
    #  ) -> RangedDescendants:
    #      really_new_descendants : RangedDescendants()#

    #        for half_move in new_descendants.range():
    #            if half_move in self:
    #               really_new_descendants_keys = set(new_descendants[half_move].keys()).difference(
    #                  set(self[half_move].keys()))
    #          else:
    #              really_new_descendants_keys = set(new_descendants[half_move].keys())
    #          for key in really_new_descendants_keys:
    #              really_new_descendants.add_descendant(new_descendants[half_move][key])
    #             self.add_descendant(new_descendants[half_move][key])

    #   # really_new_descendants.print_info()

    #    return really_new_descendants

    def merge(self, descendant_1: typing.Self, descendant_2: typing.Self) -> None:
        """
        Merges the descendants of two nodes into the current node.

        Args:
            descendant_1 (typing.Self): The first descendant node.
            descendant_2 (typing.Self): The second descendant node.

        Returns:
            None
        """
        half_moves_range = set(descendant_1.keys()) | set(descendant_2.keys())
        assert len(half_moves_range) > 0
        self.min_half_move = min(half_moves_range)
        self.max_half_move = max(half_moves_range)
        for half_move in half_moves_range:
            if descendant_1.is_in_the_current_range(half_move):
                if descendant_2.is_in_the_current_range(half_move):
                    #  print('dd',type(self.descendants_at_half_move),type())
                    # in python 3.9 we can use a |
                    self.descendants_at_half_move[half_move] = {
                        **descendant_1[half_move],
                        **descendant_2[half_move],
                    }
                    self.number_of_descendants_at_half_move[half_move] = len(
                        self[half_move]
                    )
                    assert self.number_of_descendants_at_half_move[half_move] == len(
                        {**descendant_1[half_move], **descendant_2[half_move]}
                    )
                else:
                    self.descendants_at_half_move[half_move] = descendant_1[half_move]
                    self.number_of_descendants_at_half_move[half_move] = (
                        descendant_1.number_of_descendants_at_half_move[half_move]
                    )
            else:
                self.descendants_at_half_move[half_move] = descendant_2[half_move]
                self.number_of_descendants_at_half_move[half_move] = (
                    descendant_2.number_of_descendants_at_half_move[half_move]
                )
            self.number_of_descendants += self.number_of_descendants_at_half_move[
                half_move
            ]

    def test(self) -> None:
        """
        Perform a test on the descendants object.

        This method checks the validity of the descendants object by asserting various conditions.
        If the `min_half_move` attribute is None, it asserts that `max_half_move` is also None and
        `number_of_descendants` is 0.
        Otherwise, it asserts that `max_half_move` and `min_half_move` are not None, and checks if all half moves
        between `min_half_move` and `max_half_move` are present in `descendants_at_half_move` dictionary.
        Finally, it iterates over all half moves in the descendants object and asserts that each half move
        is within the current range.

        Returns:
            None
        """
        super().test()
        if self.min_half_move is None:
            assert self.max_half_move is None
            assert self.number_of_descendants == 0
        else:
            assert self.max_half_move is not None
            assert self.min_half_move is not None
            for i in range(self.min_half_move, self.max_half_move + 1):
                assert i in self.descendants_at_half_move.keys()
        for half_move in self:
            assert self.is_in_the_current_range(half_move)

    def print_info(self) -> None:
        """
        Prints information about the descendants.

        This method calls the `print_info` method of the parent class and then prints the count of descendants,
        the minimum half move, and the maximum half move.

        Returns:
            None
        """
        super().print_info()
        print(
            "---here are the ",
            self.get_count(),
            " descendants. min:",
            self.min_half_move,
            ". max:",
            self.max_half_move,
        )


class SortedDescendants(Descendants):
    # todo is there a difference between sorted descendant nd sorted value descendant? below?

    """
    Represents a class that stores sorted descendants of a tree node at different half moves.
    Inherits from the Descendants class.
    """

    sorted_descendants_at_half_move: dict[int, dict[ITreeNode, float]]

    def __init__(self) -> None:
        super().__init__()
        self.sorted_descendants_at_half_move = {}

    def update_value(self, node: ITreeNode, value: float) -> None:
        """
        Updates the value of a descendant node.

        Args:
            node (ITreeNode): The descendant node.
            value (float): The new value for the descendant node.
        """
        self.sorted_descendants_at_half_move[node.half_move][node] = value

    def add_descendant_with_val(self, node: ITreeNode, value: float) -> None:
        """
        Adds a descendant node with its corresponding value.

        Args:
            node (ITreeNode): The descendant node to add.
            value (float): The value of the descendant node.
        """
        super().add_descendant(node)
        half_move = node.half_move

        if half_move in self.sorted_descendants_at_half_move:
            assert node not in self.sorted_descendants_at_half_move[half_move]
            self.sorted_descendants_at_half_move[half_move][node] = value
        else:  # half_move == len(self.descendants_at_half_move)
            self.sorted_descendants_at_half_move[half_move] = {node: value}

        assert self.contains_node(node)

    def test(self) -> None:
        """
        Performs a test to ensure the integrity of the data structure.
        """
        super().test()
        assert len(self.sorted_descendants_at_half_move) == len(
            self.descendants_at_half_move
        )
        assert (
            self.sorted_descendants_at_half_move.keys()
            == self.descendants_at_half_move.keys()
        )
        for half_move in self.sorted_descendants_at_half_move:
            assert len(self.sorted_descendants_at_half_move[half_move]) == len(
                self.descendants_at_half_move[half_move]
            )

    def print_info(self) -> None:
        """
        Prints information about the sorted descendants.
        """
        super().print_info()
        print("sorted")
        for half_move in self:
            print(
                "half_move: ",
                half_move,
                "| (",
                self.number_of_descendants_at_half_move[half_move],
                "descendants)",
            )  # ,                  end='| ')
            for descendant, value in self.sorted_descendants_at_half_move[
                half_move
            ].items():
                print(
                    descendant.id, descendant.fast_rep, "(" + str(value) + ")", end=" "
                )
            print("")

    def remove_descendant(self, node: ITreeNode) -> None:
        """
        Removes a descendant node from the data structure.

        Args:
            node (ITreeNode): The descendant node to remove.
        """
        super().remove_descendant(node)
        half_move = node.half_move
        self.sorted_descendants_at_half_move[half_move].pop(node)
        if half_move not in self.number_of_descendants_at_half_move:
            self.sorted_descendants_at_half_move.pop(half_move)
        assert not self.contains_node(node)

    def contains_node(self, node: ITreeNode) -> bool:
        """
        Checks if a descendant node is present in the data structure.

        Args:
            node (ITreeNode): The descendant node to check.

        Returns:
            bool: True if the descendant node is present, False otherwise.
        """
        reply_base = super().contains_node(node)
        if (
            node.half_move in self.descendants_at_half_move
            and node in self.sorted_descendants_at_half_move[node.half_move]
        ):
            rep = True
        else:
            rep = False
        assert reply_base == rep
        return rep


class SortedValueDescendants(Descendants):
    """
    Represents a class for managing sorted descendants with associated values.
    Inherits from the `Descendants` class.
    """

    sorted_descendants_at_half_move: dict[typing.Any, typing.Any]

    def __init__(self) -> None:
        """
        Initializes a Sorted Value Descendants object.
        """
        super().__init__()
        self.sorted_descendants_at_half_move = {}

    def update_value(self, node: ITreeNode, value: float) -> None:
        """
        Updates the value associated with a given node.

        Args:
            node (ITreeNode): The node to update the value for.
            value (float): The new value to associate with the node.

        Returns:
            None
        """
        self.sorted_descendants_at_half_move[node.half_move][node] = value

    def add_descendant_val(self, node: ITreeNode, value: float) -> None:
        """
        Adds a descendant node with an associated value.

        Args:
            node (ITreeNode): The descendant node to add.
            value (float): The value associated with the descendant node.

        Returns:
            None
        """
        super().add_descendant(node)
        half_move = node.half_move

        if half_move in self.sorted_descendants_at_half_move:
            assert node not in self.sorted_descendants_at_half_move[half_move]
            self.sorted_descendants_at_half_move[half_move][node] = value
        else:
            self.sorted_descendants_at_half_move[half_move] = ValueSortedDict(
                {node: value}
            )

    def test(self) -> None:
        """
        Performs a test to ensure the integrity of the sorted descendants.

        Returns:
            None
        """
        super().test()
        assert len(self.sorted_descendants_at_half_move) == len(
            self.descendants_at_half_move
        )
        assert (
            self.sorted_descendants_at_half_move.keys()
            == self.descendants_at_half_move.keys()
        )
        for half_move in self.sorted_descendants_at_half_move:
            assert len(self.sorted_descendants_at_half_move[half_move]) == len(
                self.descendants_at_half_move[half_move]
            )

    def print_info(self) -> None:
        """
        Prints information about the sorted descendants.

        Returns:
            None
        """
        super().print_info()
        print("sorted")
        for half_move in self:
            print(
                "half_move: ",
                half_move,
                "| (",
                self.number_of_descendants_at_half_move[half_move],
                "descendants)",
            )
            for descendant, value in self.sorted_descendants_at_half_move[
                half_move
            ].items():
                print(str(descendant.id) + "(" + str(value) + ")", end=" ")
            print("")

    def remove_descendant(self, node: ITreeNode) -> None:
        """
        Removes a descendant node.

        Args:
            node (ITreeNode): The descendant node to remove.

        Returns:
            None
        """
        super().remove_descendant(node)
        half_move = node.half_move

        self.sorted_descendants_at_half_move[half_move].pop(node)
        if half_move not in self.number_of_descendants_at_half_move:
            self.sorted_descendants_at_half_move.pop(half_move)



"""
This module contains the `ReplayGameScript` class, which is responsible for replaying a chess game.
"""

import os
import sys
from dataclasses import dataclass, field
from enum import Enum
from typing import Any

import dacite
import yaml
from PySide6.QtWidgets import QApplication

from chipiron.displays.gui_replay_games import MainWindow
from chipiron.environments.chess_env.board.board_chi import BoardChi
from chipiron.environments.chess_env.board.factory import create_board_chi
from chipiron.environments.chess_env.board.utils import FenPlusHistory
from chipiron.games.game.final_game_result import GameReport
from chipiron.scripts.script import Script
from chipiron.scripts.script_args import BaseScriptArgs


@dataclass
class ReplayScriptArgs:
    """
    The input arguments needed by the replay game script to run.
    """

    # path to the yaml file with the Game Report stored
    file_game_report: str = "chipiron/scripts/one_match/outputs/Sunday-10-13-2024--22:40:58:049918/games_0_W:Sequool-vs-B:Random_game_report.yaml"

    base_script_args: BaseScriptArgs = field(default_factory=BaseScriptArgs)

    # whether to display the match in a GUI
    gui: bool = False


class ReplayGameScript:
    """
    The `ReplayGameScript` class is responsible for replaying a chess game.
    """

    args_dataclass_name: type[ReplayScriptArgs] = ReplayScriptArgs

    base_script: Script[ReplayScriptArgs]
    chess_board: BoardChi

    base_experiment_output_folder = os.path.join(
        Script.base_experiment_output_folder, "replay_game/outputs/"
    )

    def __init__(
        self,
        base_script: Script[ReplayScriptArgs],
    ) -> None:
        """
        Initializes the `ReplayGameScript` object.

        Args:
            base_script (Script): The base script object.

        """
        self.base_script = base_script

        # Calling the init of Script that takes care of a lot of stuff, especially parsing the arguments into self.args
        self.args: ReplayScriptArgs = self.base_script.initiate(
            experiment_output_folder=self.base_experiment_output_folder,
        )

        with open(self.args.file_game_report, "r", encoding="utf-8") as fileGame:
            game_report_dict: dict[Any, Any] = yaml.safe_load(fileGame)
            game_report: GameReport = dacite.from_dict(
                data_class=GameReport,
                data=game_report_dict,
                config=dacite.Config(cast=[Enum]),
            )
            self.chess_board: BoardChi = create_board_chi(
                fen_with_history=FenPlusHistory(
                    current_fen=game_report.fen_history[0],
                    historical_moves=game_report.move_history,
                )
            )

    def run(self) -> None:
        """
        Runs the replay game script.

        If `gui` is set to True, it displays the match in a GUI. Otherwise, it runs a console version (TODO).

        """
        if self.args.gui:
            chess_gui = QApplication(sys.argv)
            window = MainWindow(self.chess_board)
            window.show()
            chess_gui.exec_()
        else:
            # TODO: code a console version
            ...

    def terminate(self) -> None:
        """
        Finishes the script.

        Performs any necessary cleanup or finalization steps.

        """
        self.base_script.terminate()

    @classmethod
    def get_args_dataclass_name(cls) -> type[ReplayScriptArgs]:
        """
        Returns the dataclass type that holds the arguments for the script.

        Returns:
            type: The dataclass type for the script's arguments.
        """
        return ReplayScriptArgs



"""
Module for the BoardToInput protocol and BoardToInputFunction protocol.
"""

from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

import torch

import chipiron.environments.chess_env.board as boards
from chipiron.environments.chess_env.board import IBoard
from chipiron.players.boardevaluators.neural_networks.board_to_tensor import (
    transform_board_pieces_one_side,
    transform_board_pieces_two_sides,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.board_to_transformer_input import (
    build_transformer_input,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.ModelInputRepresentationType import (
    InternalTensorRepresentationType,
    ModelInputRepresentationType,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.representation_364_bti import (
    RepresentationBTI,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.representation_factory_factory import (
    create_board_representation_factory,
)
from chipiron.players.boardevaluators.neural_networks.models.transformer_one import (
    TransformerArgs,
)

if TYPE_CHECKING:
    from chipiron.players.boardevaluators.neural_networks.input_converters.factory import (
        RepresentationFactory,
    )


@runtime_checkable
class BoardToInputFunction(Protocol):
    """
    Protocol for a callable object that converts a chess board to a tensor input for a neural network.
    """

    def __call__(self, board: boards.IBoard) -> Any:
        """
        Converts the given chess board to a tensor input.

        Args:
            board (BoardChi): The chess board to convert.

        Returns:
            torch.Tensor: The tensor input representing the chess board.
        """
        ...


class BoardToInput(Protocol):
    """
    Protocol for converting a chess board to a tensor input for a neural network.
    """

    def convert(self, board: boards.IBoard) -> torch.Tensor:
        """
        Converts the given chess board to a tensor input.

        Args:
            board (BoardChi): The chess board to convert.

        Returns:
            torch.Tensor: The tensor input representing the chess board.
        """
        ...


def create_board_to_input_from_representation(
    internal_tensor_representation_type: InternalTensorRepresentationType,
) -> BoardToInputFunction:
    """Creates a BoardToInputFunction from an InternalTensorRepresentationType.

    Args:
        internal_tensor_representation_type (InternalTensorRepresentationType): The internal tensor representation type to use.

    Returns:
        BoardToInputFunction: A function that converts a chess board to a tensor input.
    """

    representation_factory: RepresentationFactory | None = (
        create_board_representation_factory(
            internal_tensor_representation_type=internal_tensor_representation_type
        )
    )
    assert representation_factory is not None
    board_to_input_convert: BoardToInput = RepresentationBTI(
        representation_factory=representation_factory
    )
    return board_to_input_convert.convert


def create_board_to_input(
    model_input_representation_type: ModelInputRepresentationType,
) -> BoardToInputFunction:
    """Creates a BoardToInputFunction from a ModelInputRepresentationType.

    Args:
        model_input_representation_type (ModelInputRepresentationType): The model input representation type to use.

    Raises:
        Exception: If the model input representation type is not supported.

    Returns:
        BoardToInputFunction: A function that converts a chess board to a tensor input.
    """

    board_to_input_convert: BoardToInputFunction

    match model_input_representation_type:
        case ModelInputRepresentationType.BUG364:
            board_to_input_convert = create_board_to_input_from_representation(
                internal_tensor_representation_type=InternalTensorRepresentationType.BUG364
            )
        case ModelInputRepresentationType.NOBUG364:
            board_to_input_convert = create_board_to_input_from_representation(
                internal_tensor_representation_type=InternalTensorRepresentationType.NOBUG364
            )

        case ModelInputRepresentationType.PIECE_MAP:

            def board_to_input_convert_transformer(board: IBoard) -> torch.Tensor:
                return build_transformer_input(
                    piece_map=board.piece_map(),
                    board_turn=board.turn,
                    transformer_args=TransformerArgs(),
                )

            board_to_input_convert = board_to_input_convert_transformer

        case ModelInputRepresentationType.PIECE_DIFFERENCE:

            def board_to_input_convert_one_side(board: IBoard) -> torch.Tensor:
                return transform_board_pieces_one_side(board, False)

            board_to_input_convert = board_to_input_convert_one_side

        case ModelInputRepresentationType.BOARD_PIECES_TWO_SIDES:

            def board_to_input_convert_two_sides(board: IBoard) -> torch.Tensor:
                return transform_board_pieces_two_sides(board, False)

            board_to_input_convert = board_to_input_convert_two_sides

        case _:
            raise Exception(
                f"no matching case for {model_input_representation_type} in {__name__}"
            )

    return board_to_input_convert



"""
This module provides classes and functions for updating tree values in the move selector.

Classes:
- AlgorithmNodeUpdater: A class for updating algorithm nodes in the tree.
- MinMaxEvaluationUpdater: A class for updating min-max evaluation values in the tree.

Functions:
- create_algorithm_node_updater: A function for creating an instance of AlgorithmNodeUpdater.

Other:
- UpdateInstructions: A class representing update instructions for a single node.
- UpdateInstructionsBatch: A class representing a batch of update instructions.

"""

from .algorithm_node_updater import AlgorithmNodeUpdater
from .factory import create_algorithm_node_updater
from .minmax_evaluation_updater import MinMaxEvaluationUpdater
from .updates_file import (
    UpdateInstructionsFromOneNode,
    UpdateInstructionsTowardsMultipleNodes,
    UpdateInstructionsTowardsOneParentNode,
)

__all__ = [
    "create_algorithm_node_updater",
    "AlgorithmNodeUpdater",
    "UpdateInstructionsFromOneNode",
    "UpdateInstructionsTowardsOneParentNode",
    "MinMaxEvaluationUpdater",
    "UpdateInstructionsTowardsMultipleNodes",
]



from dataclasses import dataclass
from typing import Any

import chess

import chipiron.environments.chess_env.board as boards
import chipiron.players.boardevaluators.basic_evaluation as basic_evaluation
from chipiron.players.boardevaluators.all_board_evaluator_args import (
    AllBoardEvaluatorArgs,
)
from chipiron.players.boardevaluators.board_evaluator_type import BoardEvalTypes
from chipiron.players.boardevaluators.evaluation_scale import (
    EvaluationScale,
    ValueOverEnum,
    get_value_over_enum,
)
from chipiron.players.boardevaluators.neural_networks.factory import (
    create_nn_board_eval_from_nn_parameters_file_and_existing_model,
)
from chipiron.players.boardevaluators.neural_networks.neural_net_board_eval_args import (
    NeuralNetBoardEvalArgs,
)
from chipiron.players.boardevaluators.over_event import HowOver, OverEvent, Winner
from chipiron.players.boardevaluators.table_base.factory import AnySyzygyTable

from .board_evaluator import BoardEvaluator


@dataclass
class MasterBoardEvaluatorArgs:
    """
    Represents the arguments for a master board evaluator.
    """

    # Whether to use syzygy table for evaluation.
    syzygy_evaluation: bool

    # The evaluation scale used by the node evaluator. (Default values when nodes are found to be over)
    evaluation_scale: EvaluationScale

    board_evaluator: AllBoardEvaluatorArgs


class MasterBoardEvaluator:
    """
    The MasterBoardEvaluator class is responsible for evaluating the value of chess positions (that are IBoard).
    It uses a board evaluator and a syzygy evaluator to calculate the value of the positions.
    """

    # The board evaluator used to evaluate the chess board.
    board_evaluator: BoardEvaluator

    # The Syzygy table used for endgame tablebase evaluations, or None if not available.
    syzygy_evaluator: AnySyzygyTable | None

    # The value over enum used to determine the value of the node when it is over.
    value_over_enum: ValueOverEnum

    def __init__(
        self,
        board_evaluator: BoardEvaluator,
        syzygy: AnySyzygyTable | None,
        value_over_enum: ValueOverEnum,
    ) -> None:
        """
        Initializes a MasterBoardEvaluator object.

        Args:
            board_evaluator (board_evals.BoardEvaluator): The board evaluator used to evaluate the chess board.
            syzygy (SyzygyTable | None): The Syzygy table used for endgame tablebase evaluations, or None if not available.
            value_over_enum (ValueOverEnum): The value over enum used to determine the value of the node when it is over.
        """
        self.board_evaluator = board_evaluator
        self.syzygy_evaluator = syzygy
        self.value_over_enum = value_over_enum

    def value_white(self, board: boards.IBoard) -> float:
        """
        Calculates the value for the white player of a given node.
        If the value can be obtained from the syzygy evaluator, it is used.
        Otherwise, the board evaluator is used.
        """
        value_white: float | None = self.syzygy_value_white(board)
        value_white_float: float
        if value_white is None:
            value_white_float = self.board_evaluator.value_white(board)
        else:
            value_white_float = value_white
        return value_white_float

    def syzygy_value_white(self, board: boards.IBoard) -> float | None:
        """
        Calculates the value for the white player of a given board using the syzygy evaluator.
        If the syzygy evaluator is not available or the board is not in the syzygy table, None is returned.
        """
        if self.syzygy_evaluator is None or not self.syzygy_evaluator.fast_in_table(
            board
        ):
            return None
        else:
            val: int = self.syzygy_evaluator.val(board)
            return val

    def check_obvious_over_events(
        self, board: boards.IBoard
    ) -> tuple[OverEvent | None, float | None]:
        """
        Checks if the given board is in an obvious game-over state and returns the corresponding OverEvent and evaluation.

        Args:
            board (boards.IBoard): The board to evaluate for game-over conditions.

        Raises:
            ValueError: If the board result string is not recognized.

        Returns:
            tuple[OverEvent | None, float]: A tuple containing the OverEvent
            (if the game is over or can be determined from Syzygy tables, otherwise None) and the evaluation score from White's perspective.
            The evaluation is especially useful when training models.
        """
        game_over: bool = board.is_game_over()
        over_event: OverEvent | None = None
        evaluation: float | None = None
        if game_over:
            value_as_string: str = board.result(claim_draw=True)
            how_over_: HowOver
            who_is_winner_: Winner
            match value_as_string:
                case "0-1":
                    how_over_ = HowOver.WIN
                    who_is_winner_ = Winner.BLACK
                case "1-0":
                    how_over_ = HowOver.WIN
                    who_is_winner_ = Winner.WHITE
                case "1/2-1/2":
                    how_over_ = HowOver.DRAW
                    who_is_winner_ = Winner.NO_KNOWN_WINNER
                case other:
                    raise ValueError(f"value {other} not expected in {__name__}")

            over_event = OverEvent(
                how_over=how_over_,
                who_is_winner=who_is_winner_,
                termination=board.termination(),
            )

        elif self.syzygy_evaluator and self.syzygy_evaluator.fast_in_table(board):
            who_is_winner_, how_over_ = self.syzygy_evaluator.get_over_event(
                board=board
            )
            over_event = OverEvent(
                how_over=how_over_,
                who_is_winner=who_is_winner_,
                termination=None,  # not sure how to retrieve this info more precisely atm
            )
        if over_event is not None:
            evaluation = self.value_white_from_over_event(over_event=over_event)
        return over_event, evaluation

    def value_white_from_over_event(self, over_event: OverEvent) -> float:
        """
        Returns the value white given an over event.
        """
        assert over_event.is_over()
        white_value: Any
        if over_event.is_win():
            assert not over_event.is_draw()
            if over_event.is_winner(chess.WHITE):
                white_value = self.value_over_enum.VALUE_WHITE_WHEN_OVER_WHITE_WINS
            else:
                white_value = self.value_over_enum.VALUE_WHITE_WHEN_OVER_BLACK_WINS
        else:  # draw
            assert over_event.is_draw()
            white_value = self.value_over_enum.VALUE_WHITE_WHEN_OVER_DRAW
        assert isinstance(white_value, float)
        return white_value


def create_master_board_evaluator(
    board_evaluator: BoardEvaluator,
    syzygy: AnySyzygyTable | None,
    evaluation_scale: EvaluationScale,
) -> MasterBoardEvaluator:
    """
    Factory function to create a MasterBoardEvaluator instance.

    Args:
        board_evaluator (BoardEvaluator): The board evaluator to use.
        syzygy (SyzygyTable | None): The syzygy table for endgame evaluations.
        value_over_enum (ValueOverEnum): The value over enum for evaluation.

    Returns:
        MasterBoardEvaluator: An instance of MasterBoardEvaluator.
    """

    value_over_enum: ValueOverEnum = get_value_over_enum(
        evaluation_scale=evaluation_scale
    )
    return MasterBoardEvaluator(
        board_evaluator=board_evaluator, syzygy=syzygy, value_over_enum=value_over_enum
    )


def create_master_board_evaluator_from_args(
    master_board_evaluator: MasterBoardEvaluatorArgs,
    syzygy: AnySyzygyTable | None,
) -> MasterBoardEvaluator:
    """Creates a MasterBoardEvaluator instance from the given arguments.

    Args:
        master_board_evaluator (MasterBoardEvaluatorArgs): args to build the MasterBoardEvaluator
        syzygy (SyzygyTable | None): a syzygy table objecct

    Returns:
        MasterBoardEvaluator: _description_
    """
    if master_board_evaluator.syzygy_evaluation:
        syzygy_ = syzygy
    else:
        syzygy_ = None

    board_evaluator: BoardEvaluator

    match master_board_evaluator.board_evaluator.type:
        case BoardEvalTypes.BASIC_EVALUATION_EVAL:
            board_evaluator = basic_evaluation.BasicEvaluation()
        case BoardEvalTypes.NEURAL_NET_BOARD_EVAL:
            assert isinstance(
                master_board_evaluator.board_evaluator, NeuralNetBoardEvalArgs
            )
            board_evaluator = create_nn_board_eval_from_nn_parameters_file_and_existing_model(
                model_weights_file_name=master_board_evaluator.board_evaluator.neural_nets_model_and_architecture.model_weights_file_name,
                nn_architecture_args=master_board_evaluator.board_evaluator.neural_nets_model_and_architecture.nn_architecture_args,
            )
        case other:
            raise ValueError(
                f"unknown type of message received by master board evaluator {other} in {__name__}"
            )

    return create_master_board_evaluator(
        board_evaluator=board_evaluator,
        syzygy=syzygy_,
        evaluation_scale=master_board_evaluator.evaluation_scale,
    )



"""
This module contains classes for visualizing a tree structure and interacting with it.

Classes:
- PhotoViewer: A QGraphicsView widget for displaying and interacting with an image.
- Window: A QWidget that contains a PhotoViewer and provides additional functionality for tree visualization.
- VisualizeTreeScript: A script for running the tree visualization application.

Usage:
1. Create an instance of VisualizeTreeScript.
2. Call the run() method to start the visualization application.
3. Interact with the displayed tree using mouse clicks and keyboard shortcuts.
4. Call the terminate() method to finish the script.

Note: This code requires the PySide6 library and the chipiron package to be installed.
"""

import os
import pickle
import sys
import typing

from PySide6 import QtCore, QtGui, QtWidgets

from chipiron.players.move_selector.treevalue.trees.move_and_value_tree import (
    MoveAndValueTree,
)
from chipiron.players.move_selector.treevalue.trees.tree_visualization import (
    display_special,
)
from chipiron.scripts.iscript import IScript
from chipiron.scripts.script import Script


@typing.no_type_check
class PhotoViewer(QtWidgets.QGraphicsView):
    """
    A custom QGraphicsView widget for displaying photos.

    Signals:
        - photoClicked: Signal emitted when the photo is clicked. It provides the position of the click as a QPoint.

    Methods:
        - hasPhoto(): Check if the viewer has a photo loaded.
        - fitInView(scale=True): Fit the photo within the view.
        - set_photo(pixmap=None): Set the photo to be displayed.
        - wheelEvent(event): Handle the wheel event for zooming.
        - toggleDragMode(): Toggle the drag mode between ScrollHandDrag and NoDrag.
        - mousePressEvent(event): Handle the mouse press event.
        - zoomin(): Zoom in the photo.
        - zoomout(): Zoom out the photo.
    """

    photoClicked = QtCore.Signal(QtCore.QPoint)

    @typing.no_type_check
    def __init__(self, parent):
        """Initializes the PhotoViewer.

        Args:
            parent (QWidget): The parent widget.
        """
        super(PhotoViewer, self).__init__(parent)
        self._zoom = 0
        self._empty = True
        self._scene = QtWidgets.QGraphicsScene(self)
        self._photo = QtWidgets.QGraphicsPixmapItem()
        self._scene.addItem(self._photo)
        self.setScene(self._scene)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setBackgroundBrush(QtGui.QBrush(QtGui.QColor(30, 30, 30)))
        self.setFrameShape(QtWidgets.QFrame.NoFrame)

    @typing.no_type_check
    def hasPhoto(self):
        """
        Check if the viewer has a photo loaded.

        Returns:
            bool: True if a photo is loaded, False otherwise.
        """
        return not self._empty

    @typing.no_type_check
    def fitInView(self, scale=True):
        """
        Fit the photo within the view.

        Args:
            scale (bool): Whether to scale the photo to fit the view. Default is True.
        """

    @typing.no_type_check
    def set_photo(self, pixmap=None):
        """
        Set the photo to be displayed.

        Args:
            pixmap (QPixmap): The photo to be displayed. If None, the viewer will be empty.
        """
        self._zoom = 0
        if pixmap and not pixmap.isNull():
            self._empty = False
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
            self._photo.setPixmap(pixmap)
        else:
            self._empty = True
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
            self._photo.setPixmap(QtGui.QPixmap())
        self.fitInView()

    @typing.no_type_check
    def wheelEvent(self, event):
        """
        Handle the wheel event for zooming.

        Args:
            event (QWheelEvent): The wheel event object.
        """
        if self.hasPhoto():
            if event.angleDelta().y() > 0:
                factor = 1.25
                self._zoom += 1
            else:
                factor = 0.8
                self._zoom -= 1
            if self._zoom > 0:
                self.scale(factor, factor)
            elif self._zoom == 0:
                self.fitInView()
            else:
                self._zoom = 0

    @typing.no_type_check
    def toggleDragMode(self):
        """
        Toggle the drag mode between ScrollHandDrag and NoDrag.
        """
        if self.dragMode() == QtWidgets.QGraphicsView.ScrollHandDrag:
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
        elif not self._photo.pixmap().isNull():
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)

    @typing.no_type_check
    def mousePressEvent(self, event):
        """
        Handle the mouse press event.

        Args:
            event (QMouseEvent): The mouse press event object.
        """
        if self._photo.isUnderMouse():
            self.photoClicked.emit(self.mapToScene(event.pos()).toPoint())
        super(PhotoViewer, self).mousePressEvent(event)

    @typing.no_type_check
    def zoomin(self):
        """
        Zoom in the photo.
        """
        factor = 1.25
        self._zoom += 1
        if self._zoom > 0:
            self.scale(factor, factor)
        elif self._zoom == 0:
            self.fitInView()
        else:
            self._zoom = 0

    @typing.no_type_check
    def zoomout(self):
        """
        Zoom out the photo.
        """
        factor = 1 / 1.25
        self._zoom -= 1
        if self._zoom > 0:
            self.scale(factor, factor)
        elif self._zoom == 0:
            self.fitInView()
        else:
            self._zoom = 0


@typing.no_type_check
class Window(QtWidgets.QWidget):
    """
    A class representing a window for tree visualization.

    Attributes:
        viewer (PhotoViewer): The photo viewer widget.
        btnLoad (QtWidgets.QToolButton): The button to load an image.
        btnPixInfo (QtWidgets.QToolButton): The button to enter pixel info mode.
        editPixInfo (QtWidgets.QLineEdit): The line edit widget for pixel info.
        tree (MoveAndValueTree): The tree object for visualization.
        current_node (Node): The current node being displayed.
        index (dict): A dictionary mapping moves to indices.

    Methods:
        __init__(): Initializes the Window object.
        build_subtree(): Builds the subtree.
        display_subtree(): Displays the subtree.
        load_image(): Loads an image.
        pixInfo(): Toggles between drag/pan and pixel info mode.
        photoClicked(): Handles the photo clicked event.
        keyPressEvent(): Handles the key press event.
        father(): Moves to the parent node.
        move_to_son(): Moves to the specified child node.
    """

    @typing.no_type_check
    def __init__(self):
        """
        Initializes the Window class.

        This method sets up the GUI elements and initializes the necessary variables.
        It also loads the image, builds the subtree, and displays the subtree.

        Args:
            None

        Returns:
            None
        """
        super(Window, self).__init__()
        self.viewer = PhotoViewer(self)
        # 'Load image' button
        self.btnLoad = QtWidgets.QToolButton(self)
        self.btnLoad.setText("Load image")
        self.btnLoad.clicked.connect(self.load_image)
        # Button to change from drag/pan to getting pixel info
        self.btnPixInfo = QtWidgets.QToolButton(self)
        self.btnPixInfo.setText("Enter pixel info mode")
        self.btnPixInfo.clicked.connect(self.pixInfo)
        self.editPixInfo = QtWidgets.QLineEdit(self)
        self.editPixInfo.setReadOnly(True)
        self.viewer.photoClicked.connect(self.photoClicked)
        # Arrange layout
        VBlayout = QtWidgets.QVBoxLayout(self)
        VBlayout.addWidget(self.viewer)
        HBlayout = QtWidgets.QHBoxLayout()
        HBlayout.setAlignment(QtCore.Qt.AlignLeft)
        HBlayout.addWidget(self.btnLoad)
        HBlayout.addWidget(self.btnPixInfo)
        HBlayout.addWidget(self.editPixInfo)
        VBlayout.addLayout(HBlayout)

        pic = pickle.load(open("chipiron/debugTreeData_1white-#.td", "rb"))

        self.tree: MoveAndValueTree = MoveAndValueTree(
            root_node=pic[1], descendants=pic[0]
        )
        self.tree.descendants = pic[0]

        self.current_node = self.tree.root_node
        self.build_subtree()
        self.display_subtree()
        self.load_image()

    @typing.no_type_check
    def build_subtree(self):
        """
        Builds the subtree of the current node.

        This method iterates over the moves of the current node and assigns an index to each move.
        The index is stored in the `index` dictionary, where the move is the key and the index is the value.

        Returns:
            None
        """
        self.index = {}
        for ind, move in enumerate(self.current_node.moves_children):
            self.index[move] = chr(33 + ind)

    @typing.no_type_check
    def display_subtree(self):
        """
        Displays the subtree rooted at the current node.

        This method generates a visualization of the subtree rooted at the current node and saves it as a JPG image file.

        Args:
            None

        Returns:
            None
        """
        dot = display_special(
            node=self.current_node, format_str="jpg", index=self.index
        )
        dot.render("chipiron/runs/treedisplays/TreeVisualtemp")

    @typing.no_type_check
    def load_image(self):
        """
        Loads an image and sets it as the photo for the viewer.

        The image is loaded from the file 'chipiron/runs/treedisplays/TreeVisualtemp.jpg'.

        Args:
            None

        Returns:
            None
        """
        self.viewer.set_photo(
            QtGui.QPixmap("chipiron/runs/treedisplays/TreeVisualtemp.jpg")
        )

    @typing.no_type_check
    def pixInfo(self):
        """
        Toggles the drag mode of the viewer.

        This method is used to toggle the drag mode of the viewer. It switches between the modes of dragging and not dragging
        the image.

        Parameters:
            None

        Returns:
            None
        """
        self.viewer.toggleDragMode()

    @typing.no_type_check
    def photoClicked(self, pos):
        """
        Handle the event when a photo is clicked.

        Args:
            pos (QPoint): The position of the click.

        Returns:
            None
        """
        if self.viewer.dragMode() == QtWidgets.QGraphicsView.NoDrag:
            self.editPixInfo.setText("%d, %d" % (pos.x(), pos.y()))

    @typing.no_type_check
    def keyPressEvent(self, event):
        """
        Handles key press events.

        Args:
            event (QKeyEvent): The key press event.

        Returns:
            None
        """
        print(event.text())
        key = event.text()
        if key in self.index.values():
            self.move_to_son(key)
        if key == "9":
            self.viewer.zoomin()
        if key == "0":
            self.viewer.zoomout()
        if key == "z":
            self.father()

        # todo there is collision as these numbers are used for children too now...

    @typing.no_type_check
    def father(self):
        """
        Move to the parent node of the current node and perform necessary operations.
        """
        qq = list(self.current_node.parent_nodes.keys())
        father_node = qq[0]  # by default!
        if father_node is not None:
            self.current_node = father_node
            self.build_subtree()
            self.display_subtree()
            self.load_image()

    @typing.no_type_check
    def move_to_son(self, key):
        """
        Moves to the specified son node based on the given key.

        Args:
            key: The key of the son node to move to.

        Returns:
            None

        Raises:
            None
        """
        for move, ind in self.index.items():
            if key == ind:
                print("switching to move", move, ind, key)
                move_to_move = move

        self.current_node = self.current_node.moves_children[move_to_move]
        self.build_subtree()
        self.display_subtree()
        self.load_image()


class VisualizeTreeScript(IScript):
    """
    This class represents a script for visualizing a tree.
    """

    base_experiment_output_folder = os.path.join(
        Script.base_experiment_output_folder, "tree_visualization/outputs/"
    )
    base_script: Script

    def __init__(
        self,
        base_script: Script,
    ):
        """
        Initializes a TreeVisualizer object.

        Args:
            base_script (Script): The base script to visualize.

        Returns:
            None
        """

    def run(self) -> None:
        """
        Runs the tree visualizer application.
        """
        app = QtWidgets.QApplication(sys.argv)
        window = Window()
        window.setGeometry(0, 0, 1800, 1600)
        window.show()
        sys.exit(app.exec_())

    def terminate(self) -> None:
        """
        Finishing the script. Profiling or timing.
        """



"""
This module contains classes for handling datasets used in board evaluation tasks.

Classes:
- MyDataSet: A custom dataset class that loads and preprocesses data.
- FenAndValueDataSet: A subclass of MyDataSet that processes raw rows into input and target value tensors.

Functions:
- process_stockfish_value: A function that processes the stockfish value for a given board and row.
"""

import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Protocol, no_type_check

import numpy as np
import pandas
import torch
from pandas import DataFrame
from torch.utils.data import Dataset

from chipiron.environments.chess_env.board import IBoard
from chipiron.environments.chess_env.board.factory import create_board_chi
from chipiron.environments.chess_env.board.utils import FenPlusHistory, fen
from chipiron.players.boardevaluators.neural_networks.input_converters.board_to_input import (
    BoardToInputFunction,
)
from chipiron.utils import path
from chipiron.utils.logger import chipiron_logger

if TYPE_CHECKING:
    from chipiron.environments.chess_env import BoardChi


@dataclass
class DataSetArgs:
    """Arguments for the dataset."""

    train_file_name: path
    test_file_name: path | None = None
    preprocessing_data_set: bool = False


type RawSample = pandas.Series


@no_type_check
class MyDataSet[ProcessedSample](Dataset[ProcessedSample], ABC):
    """
    A custom dataset class that loads and preprocesses data.

    Attributes:
    - file_name (str): The file name of the dataset.
    - preprocessing (bool): Flag indicating whether to preprocess the dataset.
    - data (pandas.DataFrame | list[ProcessedSample] | None): The loaded and processed data.
    - len (int | None): The length of the dataset.

    Methods:
    - load(): Loads the dataset from the file.
    - process_raw_row(row: pandas.Series) -> ProcessedSample: Processes a raw row into processed sample.
    """

    data: pandas.DataFrame | list[ProcessedSample] | None
    len: int | None
    preprocessing: bool

    def __init__(self, file_name: path, preprocessing: bool) -> None:
        """
        Initializes a new instance of the MyDataSet class.

        Args:
        - file_name (str): The file name of the dataset.
        - preprocessing (bool): Flag indicating whether to preprocess the dataset.
        """
        self.file_name = file_name
        self.preprocessing = preprocessing
        self.data = None

    @no_type_check
    def load(self) -> None:
        """
        Loads the dataset from the file.
        """
        chipiron_logger.info("Loading the dataset...")
        start_time = time.time()

        raw_data_temp = pandas.read_pickle(self.file_name).copy()

        # Ensure we always work with a DataFrame
        raw_data: DataFrame
        if isinstance(raw_data_temp, pandas.Series):
            raw_data = raw_data_temp.to_frame()
        else:
            raw_data = raw_data_temp

        chipiron_logger.info("raw_data %s", str(type(raw_data).__name__))
        chipiron_logger.info(
            "--- LOAD READ PICKLE %.2f seconds ---", time.time() - start_time
        )
        chipiron_logger.info("Dataset loaded with %d items", len(raw_data))

        if self.preprocessing:
            chipiron_logger.info("Preprocessing dataset...")
            processed_data: list[ProcessedSample] = []

            # Handle DataFrame case
            for idx in range(len(raw_data)):
                if idx % 10 == 0:
                    chipiron_logger.info(
                        "Processing progress: %.2f%%", idx / len(raw_data) * 100
                    )
                row: pandas.Series[Any] = raw_data.iloc[idx]
                processed_data.append(self.process_raw_row(row))

            self.data = processed_data
            chipiron_logger.info("Preprocessing complete.")
        else:
            chipiron_logger.info("No preprocessing applied.")
            self.data = raw_data

        # Fix pandas pickle compatibility
        if isinstance(self.data, pandas.DataFrame):
            self.data = self.data.copy()
        elif isinstance(self.data, list):
            self.data = self.data.copy()

    @abstractmethod
    def process_raw_row(self, row: pandas.Series) -> ProcessedSample:
        """
        Converts a raw row into processed sample.
        Subclasses must implement this.
        """
        ...

    def __len__(self) -> int:
        """
        Returns the length of the dataset.

        Returns:
        - int: The length of the dataset.
        """
        if self.data is None:
            raise RuntimeError("Dataset not loaded yet. Call `load()` first.")
        return len(self.data)

    @no_type_check
    def __getitem__(self, idx: int) -> ProcessedSample:
        """
        Returns the item at the given index.
        Always returns a ProcessedSample, either from preprocessed data or by processing on-the-fly.

        Args:
        - idx (int): The index of the item.

        Returns:
        - ProcessedSample: The processed sample.
        """
        if self.data is None:
            raise RuntimeError("Dataset not loaded yet. Call `load()` first.")

        index = idx % len(self)
        if self.preprocessing:
            # Data is already preprocessed
            assert isinstance(self.data, list)
            return self.data[index]
        else:
            # Process on-the-fly
            assert isinstance(self.data, pandas.DataFrame)
            return self.process_raw_row(self.data.iloc[index])

    @no_type_check
    def get_unprocessed(self, idx: int) -> pandas.Series:
        """
        Returns the unprocessed raw row at the given index.

        Args:
        - idx (int): The index of the item.
        Returns:
        - pandas.Series: The raw row."""
        if not isinstance(self.data, pandas.DataFrame):
            raise RuntimeError("Unprocessed data is not available.")
        return self.data.iloc[idx % len(self)]

    def is_preprocessed(self) -> bool:
        """Checks if the dataset is preprocessed.

        Returns:
            bool: True if the dataset is preprocessed, False otherwise.
        """
        return self.preprocessing and isinstance(self.data, list)


@no_type_check
def process_stockfish_value(row: pandas.Series) -> float:
    """
    Processes the stockfish value for a given board and row.

    Args:
    - row (pandas.Series): The row from the dataset.

    Returns:
    - float: The processed target value.
    """
    # target values are value between -1 and 1 from the point of view of white. (+1 is white win and -1 is white loose)
    target_value: float = np.tanh(row["stockfish_value"] / 500.0)
    return target_value


class SupervisedData(Protocol):
    """
    A protocol that defines the structure for classes that have input and target value attributes.
    """

    is_batch: bool = False  # Flag to indicate if this contains batched data

    def get_input_layer(self) -> torch.Tensor:
        """
        Returns the input layer tensor.
        """
        ...

    def get_target_value(self) -> torch.Tensor:
        """
        Returns the target value tensor.
        """
        ...


@dataclass
class FenAndValueData:
    """Represents the FEN and value data for a chess position."""

    fen_tensor: torch.Tensor
    value_tensor: torch.Tensor
    is_batch: bool = False  # Flag to indicate if this contains batched data

    def get_input_layer(self) -> torch.Tensor:
        """
        Returns the input layer tensor.
        """
        return self.fen_tensor

    def get_target_value(self) -> torch.Tensor:
        """
        Returns the target value tensor.
        """
        return self.value_tensor


def custom_collate_fn_fen_and_value(batch: list[FenAndValueData]) -> FenAndValueData:
    inputs = [item.get_input_layer() for item in batch]
    targets = [item.get_target_value() for item in batch]

    inputs_batch = torch.stack(inputs)
    targets_batch = torch.stack(targets)

    return FenAndValueData(
        fen_tensor=inputs_batch, value_tensor=targets_batch, is_batch=True
    )


class FenAndValueDataSet(MyDataSet[FenAndValueData]):
    """
    A subclass of MyDataSet that processes raw rows into input and target value tensors.

    Attributes:
    - transform_board_function (BoardToInputFunction): The function to transform the board into input tensor.
    - transform_value_function (callable): The function to transform the value for a given board and row.

    Methods:
    - process_raw_row(row: pandas.Series) -> FenAndValueData: Processes a raw row into FenAndValueData.
    - process_raw_rows(dataframe: pandas.DataFrame) -> list[FenAndValueData]: Processes raw rows into FenAndValueData list.
    """

    transform_board_function: BoardToInputFunction  # transform board to model input
    transform_dataset_value_to_white_value_function: Callable[
        [pandas.Series], float
    ]  # transform value in dataset to standardized value white float
    transform_white_value_to_model_output_function: Callable[
        [float, IBoard], torch.Tensor
    ]  # transform white value to model output

    def __init__(
        self,
        file_name: path,
        transform_white_value_to_model_output_function: Callable[
            [float, IBoard], torch.Tensor
        ],
        transform_dataset_value_to_white_value_function: Callable[
            [pandas.Series], float
        ],
        transform_board_function: BoardToInputFunction,
        preprocessing: bool = False,
    ) -> None:
        """
        Initializes a new instance of the FenAndValueDataSet class.

        Args:
        - file_name (str): The file name of the dataset.
        - preprocessing (bool): Flag indicating whether to preprocess the dataset.
        - transform_board_function (BoardToInputFunction): The function to transform the board into input tensor.
        - transform_dataset_value_to_white_value_function: The function to transform dataset values to white values.
        - transform_white_value_to_model_output_function: The function to transform white values to model output.
        """
        super().__init__(file_name, preprocessing)

        assert isinstance(transform_board_function, BoardToInputFunction)
        self.transform_board_function = transform_board_function

        # transform function
        self.transform_dataset_value_to_white_value_function = (
            transform_dataset_value_to_white_value_function
        )

        self.transform_white_value_to_model_output_function = (
            transform_white_value_to_model_output_function
        )

    @no_type_check
    def process_raw_row(self, row: pandas.Series) -> FenAndValueData:
        """
        Processes a raw row into FenAndValueData.

        Args:
        - row (pandas.Series): The raw row from the dataset.

        Returns:
        - FenAndValueData: The processed data containing input and target tensors.
        """
        fen_: fen = row["fen"]

        # todo should we make it general and allow rust boards just for testing all comptabilities ?
        board: BoardChi = create_board_chi(
            fen_with_history=FenPlusHistory(current_fen=fen_),
            use_board_modification=True,
        )
        input_layer: torch.Tensor = self.transform_board_function(board)
        target_value_white: float = (
            self.transform_dataset_value_to_white_value_function(row)
        )
        target_value: torch.Tensor = (
            self.transform_white_value_to_model_output_function(
                target_value_white, board
            )
        )
        return FenAndValueData(fen_tensor=input_layer, value_tensor=target_value)

    def process_raw_rows(self, dataframe: pandas.DataFrame) -> list[FenAndValueData]:
        """
        Processes raw rows into input and target tensors.

        Args:
        - dataframe (pandas.DataFrame): The raw rows from the dataset.

        Returns:
        - list[FenAndValueData]: The processed input and target tensors.
        """
        processed_data: list[FenAndValueData] = []
        row: pandas.Series
        for _, row in dataframe.iterrows():
            processed_data.append(self.process_raw_row(row))

        return processed_data



import shakmaty_python_binding

from .utils import moveUci


class RustMove:
    move: shakmaty_python_binding.MyMove
    uci_: moveUci

    def __init__(self, move: shakmaty_python_binding.MyMove, uci: moveUci) -> None:
        self.move = move
        self.uci_ = uci

    def is_zeroing(self) -> bool:
        return self.move.is_zeroing()

    def uci(self) -> moveUci:
        return self.uci_



"""
This module provides functionality for creating and managing Sequool objects.

A Sequool object represents a sequence of operations that can be applied to a dataset.

Example usage:
    from .factory import create_sequool, SequoolArgs

    # Create a Sequool object
    sequool = create_sequool()

    # Define the arguments for the Sequool object
    sequool_args = SequoolArgs()

"""

from .factory import SequoolArgs, create_sequool

__all__ = ["create_sequool", "SequoolArgs"]



"""
Module collecting the progress of computing moves by each player
"""

import queue
from dataclasses import dataclass, field
from typing import Protocol

import chess

from chipiron.utils.communication.gui_messages import PlayerProgressMessage
from chipiron.utils.dataclass import IsDataclass


class PlayerProgressCollectorP(Protocol):
    """
    Object defining the protocol for setting the progress values
    """

    def progress_white(self, value: int | None) -> None: ...

    def progress_black(self, value: int | None) -> None: ...


@dataclass
class PlayerProgressCollector:
    """
    Object in charge of collecting the progress of computing moves by each player
    """

    progress_white_: int | None = None
    progress_black_: int | None = None

    @property
    def progress_white(self) -> int | None:
        """Gets the progress of the white player.

        Returns:
            int | None: The progress of the white player.
        """
        return self.progress_white_

    @progress_white.setter
    def progress_white(self, value: int | None) -> None:
        """Sets the progress of the white player.

        Args:
            value (int | None): The progress of the white player.
        """
        self.progress_white_ = value

    @property
    def progress_black(self) -> int | None:
        """Gets the progress of the black player.

        Returns:
            int | None: The progress of the black player.
        """
        return self.progress_black_

    @progress_black.setter
    def progress_black(self, value: int | None) -> None:
        """Sets the progress of the black player.

        Args:
            value (int | None): The progress of the black player.
        """
        self.progress_black_ = value


@dataclass
class PlayerProgressCollectorObservable:
    """
    Object in charge of collecting the progress of computing moves by each player
    """

    progress_collector: PlayerProgressCollector = field(
        default_factory=PlayerProgressCollector
    )
    subscribers: list[queue.Queue[IsDataclass]] = field(
        default_factory=lambda: list[queue.Queue[IsDataclass]]()
    )

    def progress_white(self, value: int | None) -> None:
        """Sets the progress of the white player and notifies subscribers."""
        self.progress_collector.progress_white = value
        self.notify(color=chess.WHITE, value=value)

    def progress_black(self, value: int | None) -> None:
        """Sets the progress of the black player and notifies subscribers."""
        self.progress_collector.progress_black = value
        self.notify(color=chess.BLACK, value=value)

    def notify(self, color: chess.Color, value: int | None) -> None:
        """Notifies subscribers of the progress of the given player color."""
        for subscriber in self.subscribers:
            subscriber.put(
                PlayerProgressMessage(player_color=color, progress_percent=value)
            )



"""
Module that contains the logic to compute the exploration index of a node in a tree.
"""

import math
from typing import TYPE_CHECKING, Protocol

import chess

import chipiron.players.move_selector.treevalue.trees as trees
from chipiron.players.move_selector.treevalue.indices.node_indices.index_data import (
    IntervalExplo,
    MinMaxPathValue,
    RecurZipfQuoolExplorationData,
)
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)
from chipiron.utils.small_tools import (
    Interval,
    distance_number_to_interval,
    intersect_intervals,
)

if TYPE_CHECKING:
    from chipiron.environments.chess_env.move.imove import moveKey
    from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode


class NodeExplorationIndexManager(Protocol):
    """
    A protocol for managing the exploration indices of nodes in a tree.

    This protocol defines methods for updating the exploration indices of nodes in a tree.

    Args:
        Protocol (type): The base protocol type.
    """

    def update_root_node_index(
        self,
        root_node: AlgorithmNode,
    ) -> None:
        """
        Updates the exploration index of the root node in the tree.

        Args:
            root_node (AlgorithmNode): The root node of the tree.
        """
        ...

    def update_node_indices(
        self,
        child_node: AlgorithmNode,
        parent_node: AlgorithmNode,
        tree: trees.MoveAndValueTree,
        child_rank: int,
    ) -> None:
        """
        Updates the exploration index of a child node in the tree.

        Args:
            child_node (AlgorithmNode): The child node to update.
            parent_node (AlgorithmNode): The parent node of the child node.
            tree (trees.MoveAndValueTree): The tree containing the nodes.
            child_rank (int): The rank of the child node among its siblings.
        """
        ...


class NullNodeExplorationIndexManager(NodeExplorationIndexManager):
    """
    A class representing a null node exploration index manager.

    This class is used when there is no need to update the exploration index of nodes in a tree.
    It inherits from the NodeExplorationIndexManager class.
    """

    def update_root_node_index(
        self,
        root_node: AlgorithmNode,
    ) -> None:
        """
        Updates the exploration index of the root node in the tree.

        Args:
            root_node (AlgorithmNode): The root node of the tree.
        """
        ...

    def update_node_indices(
        self,
        child_node: AlgorithmNode,
        parent_node: AlgorithmNode,
        tree: trees.MoveAndValueTree,
        child_rank: int,
    ) -> None:
        """
        Updates the exploration index of a child node in the tree.

        Args:
            child_node (AlgorithmNode): The child node to update.
            parent_node (AlgorithmNode): The parent node of the child node.
            tree (trees.MoveAndValueTree): The tree containing the nodes.
            child_rank (int): The rank of the child node among its siblings.
        """
        raise Exception("should not be raised")


class UpdateIndexGlobalMinChange:
    """
    A class that updates the exploration index of nodes in a tree using the global minimum change strategy.
    """

    def update_root_node_index(
        self,
        root_node: AlgorithmNode,
    ) -> None:
        """
        Updates the exploration index of the root node in the tree using the global minimum change strategy.

        Args:
            root_node (AlgorithmNode): The root node of the tree.
        """
        assert isinstance(root_node.exploration_index_data, MinMaxPathValue)
        root_value: float = root_node.minmax_evaluation.get_value_white()

        root_node.exploration_index_data.min_path_value = root_value
        root_node.exploration_index_data.max_path_value = root_value
        root_node.exploration_index_data.index = 0

    def update_node_indices(
        self,
        child_node: AlgorithmNode,
        parent_node: AlgorithmNode,
        tree: trees.MoveAndValueTree,
        child_rank: int,
    ) -> None:
        """
        Updates the exploration index of a child node in the tree using the global minimum change strategy.

        Args:
            child_node (AlgorithmNode): The child node to update.
            parent_node (AlgorithmNode): The parent node of the child node.
            tree (trees.MoveAndValueTree): The tree containing the nodes.
            child_rank (int): The rank of the child node among its siblings.
        """
        assert isinstance(parent_node.exploration_index_data, MinMaxPathValue)
        assert isinstance(child_node.exploration_index_data, MinMaxPathValue)

        assert parent_node.exploration_index_data is not None
        assert parent_node.exploration_index_data.min_path_value is not None
        assert parent_node.exploration_index_data.max_path_value is not None
        assert child_node.exploration_index_data is not None

        child_value: float = child_node.minmax_evaluation.get_value_white()

        child_node_min_path_value = min(
            child_value, parent_node.exploration_index_data.min_path_value
        )

        child_node_max_path_value = max(
            child_value, parent_node.exploration_index_data.max_path_value
        )

        # computes local_child_index the amount of change for the child node to become better than its parent
        child_index: float = (
            abs(child_node_max_path_value - child_node_min_path_value) / 2
        )

        # the amount of change for the child to become better than any of its ancestor
        # and become the overall best bode, the max is computed with the parent index
        # child_index: float = max(local_child_index, parent_index)

        # the index of the child node is updated now
        # as a child node can have multiple parents we take the min if an index was previously computed
        if child_node.exploration_index_data.index is None:
            child_node.exploration_index_data.index = child_index
            child_node.exploration_index_data.max_path_value = child_node_max_path_value
            child_node.exploration_index_data.min_path_value = child_node_min_path_value
        else:
            assert child_node.exploration_index_data.max_path_value is not None
            assert child_node.exploration_index_data.min_path_value is not None
            child_node.exploration_index_data.index = min(
                child_node.exploration_index_data.index, child_index
            )
            child_node.exploration_index_data.max_path_value = min(
                child_node_max_path_value,
                child_node.exploration_index_data.max_path_value,
            )
            child_node.exploration_index_data.min_path_value = max(
                child_node_min_path_value,
                child_node.exploration_index_data.min_path_value,
            )


class UpdateIndexZipfFactoredProba:
    """
    A class that updates the exploration index of nodes in a tree using the Zipf factored probability strategy.
    """

    def update_root_node_index(
        self,
        root_node: AlgorithmNode,
    ) -> None:
        """
        Updates the exploration index of the root node in the tree using the Zipf factored probability strategy.

        Args:
            root_node (AlgorithmNode): The root node of the tree.
        """
        assert isinstance(
            root_node.exploration_index_data, RecurZipfQuoolExplorationData
        )

        root_node.exploration_index_data.zipf_factored_proba = 1
        root_node.exploration_index_data.index = 0

    def update_node_indices(
        self,
        child_node: AlgorithmNode,
        parent_node: AlgorithmNode,
        tree: trees.MoveAndValueTree,
        child_rank: int,
    ) -> None:
        """
        Updates the exploration index of a child node in the tree using the Zipf factored probability strategy.

        Args:
            child_node (AlgorithmNode): The child node to update.
            parent_node (AlgorithmNode): The parent node of the child node.
            tree (trees.MoveAndValueTree): The tree containing the nodes.
            child_rank (int): The rank of the child node among its siblings.
        """
        assert isinstance(
            parent_node.exploration_index_data, RecurZipfQuoolExplorationData
        )

        parent_zipf_factored_proba: float | None = (
            parent_node.exploration_index_data.zipf_factored_proba
        )
        assert parent_zipf_factored_proba is not None

        child_zipf_proba: float = 1 / (child_rank + 1)
        child_zipf_factored_proba: float = child_zipf_proba * parent_zipf_factored_proba
        inverse_depth: float = 1 / (tree.node_depth(child_node) + 1)
        child_index: float = child_zipf_factored_proba * inverse_depth
        child_index = -child_index

        assert child_node.exploration_index_data is not None
        assert isinstance(
            child_node.exploration_index_data, RecurZipfQuoolExplorationData
        )

        # the index of the child node is updated now
        # as a child node can have multiple parents we take the min if an index was previously computed
        if child_node.exploration_index_data.index is None:
            child_node.exploration_index_data.index = child_index
            child_node.exploration_index_data.zipf_factored_proba = (
                child_zipf_factored_proba
            )

        else:
            assert child_node.exploration_index_data.zipf_factored_proba is not None
            child_node.exploration_index_data.index = min(
                child_node.exploration_index_data.index, child_index
            )
            child_node.exploration_index_data.zipf_factored_proba = min(
                child_node.exploration_index_data.zipf_factored_proba,
                child_zipf_factored_proba,
            )


class UpdateIndexLocalMinChange:
    """
    A class that updates the exploration index of nodes in a tree using the local minimum change strategy.
    """

    def update_root_node_index(
        self,
        root_node: AlgorithmNode,
    ) -> None:
        """
        Updates the exploration index of the root node in the tree using the local minimum change strategy.

        Args:
            root_node (AlgorithmNode): The root node of the tree.
        """
        assert isinstance(root_node.exploration_index_data, IntervalExplo)

        root_node.exploration_index_data.index = 0
        root_node.exploration_index_data.interval = Interval(
            min_value=-math.inf, max_value=math.inf
        )

    def update_node_indices(
        self,
        child_node: AlgorithmNode,
        parent_node: AlgorithmNode,
        tree: trees.MoveAndValueTree,
        child_rank: int,
    ) -> None:
        """
        Updates the exploration index of a child node in the tree using the local minimum change strategy.

        Args:
            child_node (AlgorithmNode): The child node to update.
            parent_node (AlgorithmNode): The parent node of the child node.
            tree (trees.MoveAndValueTree): The tree containing the nodes.
            child_rank (int): The rank of the child node among its siblings.
        """
        assert isinstance(parent_node.exploration_index_data, IntervalExplo)

        assert parent_node.exploration_index_data is not None
        assert child_node.exploration_index_data is not None

        inter_level_interval: Interval | None = None
        local_index = None

        if parent_node.exploration_index_data.index is None:
            child_node.exploration_index_data.index = None
        else:
            assert parent_node.exploration_index_data.interval is not None
            if len(parent_node.tree_node.moves_children) == 1:
                local_index = parent_node.exploration_index_data.index
                inter_level_interval = parent_node.exploration_index_data.interval
            else:
                if parent_node.tree_node.board.turn == chess.WHITE:
                    best_move: moveKey | None = (
                        parent_node.minmax_evaluation.best_move()
                    )
                    assert best_move is not None
                    best_child = parent_node.moves_children[best_move]
                    assert isinstance(best_child, AlgorithmNode)
                    second_best_move: moveKey | None = (
                        parent_node.minmax_evaluation.second_best_move()
                    )
                    assert second_best_move is not None
                    second_best_child = parent_node.moves_children[second_best_move]
                    child_white_value = child_node.minmax_evaluation.get_value_white()
                    local_interval = Interval()
                    if child_node == best_child:
                        assert isinstance(second_best_child, AlgorithmNode)
                        local_interval.max_value = math.inf
                        local_interval.min_value = (
                            second_best_child.minmax_evaluation.get_value_white()
                        )
                    else:
                        local_interval.max_value = math.inf
                        local_interval.min_value = (
                            best_child.minmax_evaluation.get_value_white()
                        )

                    inter_level_interval = intersect_intervals(
                        local_interval, parent_node.exploration_index_data.interval
                    )
                    if inter_level_interval is not None:
                        local_index = distance_number_to_interval(
                            value=child_white_value, interval=inter_level_interval
                        )
                    else:
                        ...
                        local_index = None
                if parent_node.tree_node.board.turn == chess.BLACK:
                    best_move_black: moveKey | None = (
                        parent_node.minmax_evaluation.best_move()
                    )
                    assert best_move_black is not None
                    best_child = parent_node.moves_children[best_move_black]
                    second_best_move_black: moveKey | None = (
                        parent_node.minmax_evaluation.second_best_move()
                    )
                    assert second_best_move_black is not None
                    second_best_child = parent_node.moves_children[
                        second_best_move_black
                    ]
                    child_white_value = child_node.minmax_evaluation.get_value_white()
                    local_interval = Interval()
                    assert isinstance(best_child, AlgorithmNode)
                    if child_node == best_child:
                        assert isinstance(second_best_child, AlgorithmNode)
                        local_interval.max_value = (
                            second_best_child.minmax_evaluation.get_value_white()
                        )
                        local_interval.min_value = -math.inf
                    else:
                        local_interval.max_value = (
                            best_child.minmax_evaluation.get_value_white()
                        )
                        local_interval.min_value = -math.inf

                    inter_level_interval = intersect_intervals(
                        local_interval, parent_node.exploration_index_data.interval
                    )
                    if inter_level_interval is not None:
                        local_index = distance_number_to_interval(
                            value=child_white_value, interval=inter_level_interval
                        )
                    else:
                        local_index = None

            assert isinstance(child_node.exploration_index_data, IntervalExplo)

            if child_node.exploration_index_data.index is None:
                child_node.exploration_index_data.index = local_index
                child_node.exploration_index_data.interval = inter_level_interval

            elif local_index is not None:
                if local_index < child_node.exploration_index_data.index:
                    child_node.exploration_index_data.interval = inter_level_interval
                child_node.exploration_index_data.index = min(
                    child_node.exploration_index_data.index, local_index
                )


# TODO their might be ways to optimize the computation such as not recomptuing for the whole tree
def update_all_indices(
    tree: trees.MoveAndValueTree, index_manager: NodeExplorationIndexManager
) -> None:
    """
    The idea is to compute an index $ind(n)$ for a node $n$ that measures the minimum amount of change
     in the value of all the nodes such that this node $n$ becomes the best.

    This can be computed recursively as :
    ind(n) = max( ind(parent(n),.5*abs(value(n)-value(parent(n))))

    Args:
        index_manager:
        tree: a tree

    Returns:

    """
    if isinstance(index_manager, NullNodeExplorationIndexManager):
        return

    tree_nodes: trees.RangedDescendants = tree.descendants

    index_manager.update_root_node_index(
        root_node=tree.root_node,
    )

    half_move: int
    for half_move in tree_nodes:
        # todo how are we sure that the hm comes in order?
        parent_node: ITreeNode
        for parent_node in tree_nodes[half_move].values():
            assert isinstance(parent_node, AlgorithmNode)
            child_node: ITreeNode | None
            # for child_node in parent_node.moves_children.values():
            move_rank: int
            move: moveKey
            for move_rank, move in enumerate(
                parent_node.minmax_evaluation.moves_sorted_by_value_
            ):
                child_node = parent_node.moves_children[move]
                assert isinstance(child_node, AlgorithmNode)
                index_manager.update_node_indices(
                    child_node=child_node,
                    tree=tree,
                    child_rank=move_rank,
                    parent_node=parent_node,
                )


# TODO their might be ways to optimize the computation such as not recomptuing for the whole tree


def print_all_indices(
    tree: trees.MoveAndValueTree,
) -> None:
    """
    Prints the exploration indices of all nodes in the given tree.

    Args:
        tree (trees.MoveAndValueTree): The tree containing the nodes.

    Returns:
        None
    """
    tree_nodes: trees.RangedDescendants = tree.descendants

    half_move: int
    for half_move in tree_nodes:
        # todo how are we sure that the hm comes in order?
        parent_node: ITreeNode
        for parent_node in tree_nodes[half_move].values():
            assert isinstance(parent_node, AlgorithmNode)
            if parent_node.exploration_index_data is not None:
                print(
                    "parent_node",
                    parent_node.tree_node.id,
                    parent_node.exploration_index_data.index,
                )



"""
the one match script
"""

import multiprocessing
import os
import sys
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, cast

import yaml
from PySide6.QtWidgets import QApplication

import chipiron as ch
import chipiron.displays as display
from chipiron.environments.chess_env.board import BoardFactory, create_board_factory
from chipiron.games.match.match_args import MatchArgs
from chipiron.games.match.match_factories import create_match_manager_from_args
from chipiron.scripts.chipiron_args import ImplementationArgs
from chipiron.scripts.script import Script
from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils.logger import chipiron_logger

if TYPE_CHECKING:
    import queue

    from chipiron.utils.dataclass import IsDataclass


@dataclass
class MatchScriptArgs:
    """
    The input arguments needed by the one match script to run
    """

    base_script_args: BaseScriptArgs = field(default_factory=BaseScriptArgs)
    match_args: MatchArgs = field(default_factory=MatchArgs)
    implementation_args: ImplementationArgs = field(default_factory=ImplementationArgs)

    # whether to display the match in a GUI
    gui: bool = True


class OneMatchScript:
    """
    Script that plays a match between two players

    """

    args_dataclass_name: type[MatchScriptArgs] = MatchScriptArgs

    base_experiment_output_folder = os.path.join(
        Script.base_experiment_output_folder, "one_match/outputs/"
    )
    base_script: Script[MatchScriptArgs]

    chess_gui: QApplication

    process_match_manager: multiprocessing.Process

    def __init__(
        self,
        base_script: Script[MatchScriptArgs],
    ) -> None:
        """
        Builds the OneMatchScript object
        """

        self.base_script = base_script

        # Calling the init of Script that takes care of a lot of stuff, especially parsing the arguments into args
        args: MatchScriptArgs = self.base_script.initiate(
            experiment_output_folder=self.base_experiment_output_folder,
        )

        # creating the match manager
        self.match_manager: ch.game.MatchManager = create_match_manager_from_args(
            match_args=args.match_args,
            base_script_args=args.base_script_args,
            implementation_args=args.implementation_args,
            gui=args.gui,
        )

        # saving the arguments of the script
        with open(
            os.path.join(
                str(args.base_script_args.experiment_output_folder),
                "inputs_and_parsing/one_match_script_merge.yaml",
            ),
            "w",
            encoding="utf-8",
        ) as one_match_script:
            yaml.dump(asdict(args), one_match_script, default_flow_style=False)

        # checking for some incompatibility
        if args.gui and args.base_script_args.profiling:
            raise ValueError("Profiling does not work well atm with gui on")

        # If we need a GUI
        if args.gui:
            # if we use a graphic user interface (GUI) we create it its own thread and
            # create its mailbox to communicate with other threads
            gui_thread_mailbox: queue.Queue[IsDataclass] = (
                multiprocessing.Manager().Queue()
            )

            app = QApplication.instance()
            if app is None:
                app = QApplication(sys.argv)
            else:
                app = cast("QApplication", app)
            self.chess_gui = app

            board_factory: BoardFactory = create_board_factory(
                use_rust_boards=args.implementation_args.use_rust_boards,
                sort_legal_moves=args.base_script_args.universal_behavior,
            )
            self.window: display.MainWindow = display.MainWindow(
                gui_mailbox=gui_thread_mailbox,
                main_thread_mailbox=self.match_manager.game_manager_factory.main_thread_mailbox,
                board_factory=board_factory,
            )
            self.match_manager.subscribe(gui_thread_mailbox)

        self.gui = args.gui

    def run(self) -> None:
        """
        Runs the match either with a GUI or not
        """

        chipiron_logger.info(" Script One Match go")
        # Qt Application needs to be in the main Thread, so we need to distinguish between GUI and no GUI
        if self.gui:  # case with GUI
            # Launching the Match Manager in a Thread
            self.process_match_manager = multiprocessing.Process(
                target=self.match_manager.play_one_match
            )
            self.process_match_manager.start()

            # Qt Application launched in the main thread
            self.window.show()
            self.chess_gui.exec_()
        else:  # No GUI
            self.match_manager.play_one_match()

        chipiron_logger.info("Finish the run of the match")

        # TODO check the good closing of processes

    def terminate(self) -> None:
        """
        Terminates the script and cleans up any resources.
        """
        chipiron_logger.info("terminating script")
        self.base_script.terminate()
        if self.gui:
            self.process_match_manager.terminate()

    @classmethod
    def get_args_dataclass_name(cls) -> type[MatchScriptArgs]:
        """
        Returns the dataclass type that holds the arguments for the script.

        Returns:
            type: The dataclass type for the script's arguments.
        """
        return MatchScriptArgs



"""
This module defines the MatchConfigTag enumeration for referencing specific match configuration YAML files
in the 'data/settings/OneMatch' directory. Each tag corresponds to a configuration file and provides methods
to retrieve the file path and parse its contents into a MatchSettingsArgs dataclass.

Classes:
    MatchConfigTag (Enum): Enum representing available match configuration tags.

Methods:
    get_yaml_file_path() -> path:
        Returns the file path to the YAML configuration file associated with the tag.

    get_match_settings_args() -> MatchSettingsArgs:
        Parses the YAML configuration file and returns its contents as a MatchSettingsArgs instance.
"""

from enum import Enum
from importlib.resources import as_file, files

import parsley_coco

from chipiron.games.match.match_settings_args import MatchSettingsArgs
from chipiron.utils import path


class MatchConfigTag(str, Enum):
    """
    Enum representing available match configuration tags.
    """

    # this list should correspond to files existing in the data/settings/OneMatch folder

    CUBO = "setting_cubo"
    DUDA = "setting_duda"
    JIME = "setting_jime"
    TRON = "setting_tron"

    def get_yaml_file_path(self) -> path:
        """Returns the file path to the YAML configuration file associated with the tag.

        Returns:
            path: The file path to the YAML configuration file.
        """
        resource = files("chipiron").joinpath(
            "data/settings/OneMatch", self.value + ".yaml"
        )
        with as_file(resource) as path_player:
            return path_player

    def get_match_settings_args(self) -> MatchSettingsArgs:
        """Parses the YAML configuration file and returns its contents as a MatchSettingsArgs instance.

        Returns:
            MatchSettingsArgs: The parsed match settings.
        """
        match_args: MatchSettingsArgs = (
            parsley_coco.resolve_yaml_file_to_base_dataclass(
                yaml_path=str(self.get_yaml_file_path()),
                base_cls=MatchSettingsArgs,
                raise_error_with_nones=False,
            )
        )
        return match_args



"""
This module provides functionality for creating and using table bases for board evaluation.

Table bases are precomputed endgame databases that store optimal moves for every possible position in a specific
 endgame scenario. These databases can be used to improve the performance of board evaluation algorithms
  by providing accurate evaluations for endgame positions.

This module exports the following functions and classes:
- create_syzygy: A function for creating a SyzygyTable object.
- SyzygyTable: A class representing a table base for endgame evaluation.
"""

from .syzygy_table import SyzygyTable

__all__ = ["SyzygyTable"]



"""
This script learns a neural network (NN) from a supervised dataset of board and evaluation pairs.

The script takes care of setting up the data loader from the evaluation files, creating the NN,
and running the learning process.

Usage:
- Instantiate the `LearnNNScript` class with a `base_script` object.
- Call the `run()` method to start the learning process.
- Call the `terminate()` method to finish the script.

Example:
    base_script = Script()
    learn_script = LearnNNScript(base_script)
    learn_script.run()
    learn_script.terminate()
"""

import copy
import logging
import os
import tempfile
import time
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any

import mlflow
import mlflow.pytorch
import torch
from mlflow.models.signature import (
    ModelSignature,
    infer_signature,  # pyright: ignore[reportUnknownVariableType]
)
from torch.utils.data import DataLoader
from torchinfo import summary  # pyright: ignore[reportUnknownVariableType]

import chipiron.utils.path_variables
from chipiron.learningprocesses.nn_trainer.factory import (
    NNTrainerArgs,
    create_nn_trainer,
    safe_nn_architecture_save,
    safe_nn_param_save,
    safe_nn_trainer_save,
)
from chipiron.players.boardevaluators.datasets.datasets import (
    DataSetArgs,
    FenAndValueData,
    FenAndValueDataSet,
    custom_collate_fn_fen_and_value,
    process_stockfish_value,  # pyright: ignore[reportUnknownVariableType]
)
from chipiron.players.boardevaluators.neural_networks import NNBoardEvaluator
from chipiron.players.boardevaluators.neural_networks.factory import (
    create_nn_board_eval_from_architecture_args,
    create_nn_board_eval_from_nn_parameters_file_and_existing_model,
)
from chipiron.scripts.script import Script
from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils import path
from chipiron.utils.chi_nn import ChiNN
from chipiron.utils.logger import chipiron_logger

if TYPE_CHECKING:
    from chipiron.learningprocesses.nn_trainer.nn_trainer import NNPytorchTrainer

logging.basicConfig(level=logging.WARNING)


@dataclass
class LearnNNScriptArgs:
    """
    Represents the arguments for the LearnNNScript.

    Attributes:
        nn_trainer_args (NNTrainerArgs): The arguments for the NNTrainer.

    """

    base_script_args: BaseScriptArgs = field(default_factory=BaseScriptArgs)
    nn_trainer_args: NNTrainerArgs = field(default_factory=NNTrainerArgs)
    dataset_args: DataSetArgs = field(
        default_factory=lambda: DataSetArgs(
            train_file_name="data/datasets/goodgames_plusvariation_stockfish_eval_train_t.1_merge.pi",
            test_file_name="data/datasets/goodgames_plusvariation_stockfish_eval_test",
            preprocessing_data_set=False,
        )
    )


class LearnNNScript:
    """
    Script that learns a NN from a supervised dataset pairs of board and evaluation

    Args:
        base_script (Script): The base script object.

    Attributes:
        base_script (Script): The base script object.
        nn (ChiNN): The neural network object.
        args (LearnNNScriptArgs): The script arguments.

    """

    args_dataclass_name: type[LearnNNScriptArgs] = LearnNNScriptArgs

    base_experiment_output_folder = os.path.join(
        Script.base_experiment_output_folder,
        "learn_nn_supervised/learn_nn_supervised_outputs",
    )

    base_script: Script[LearnNNScriptArgs]
    nn: ChiNN
    args: LearnNNScriptArgs
    nn_board_evaluator: NNBoardEvaluator
    saving_folder: path

    def __init__(
        self,
        base_script: Script[LearnNNScriptArgs],
    ) -> None:
        """
        Initializes the LearnNNFromSupervisedDatasets class.

        Args:
            base_script (Script): The base script object.

        Returns:
            None
        """

        self.start_time = time.time()

        # Setting up the dataloader from the evaluation files
        self.base_script = base_script

        # Calling the init of Script that takes care of a lot of stuff, especially parsing the arguments into self.args
        self.args: LearnNNScriptArgs = self.base_script.initiate(
            experiment_output_folder=self.base_experiment_output_folder,
        )

        # taking care of random
        chipiron.set_seeds(seed=self.args.base_script_args.seed)

        if self.args.nn_trainer_args.reuse_existing_model:
            assert (
                self.args.nn_trainer_args.nn_parameters_file_if_reusing_existing_one
                is not None
            )
            self.nn_board_evaluator = create_nn_board_eval_from_nn_parameters_file_and_existing_model(
                model_weights_file_name=self.args.nn_trainer_args.nn_parameters_file_if_reusing_existing_one,
                nn_architecture_args=self.args.nn_trainer_args.neural_network_architecture_args,
            )
        else:
            self.nn_board_evaluator = create_nn_board_eval_from_architecture_args(
                nn_architecture_args=self.args.nn_trainer_args.neural_network_architecture_args
            )

        if self.args.nn_trainer_args.specific_saving_folder is not None:
            self.saving_folder = self.args.nn_trainer_args.specific_saving_folder
        else:
            assert self.args.base_script_args.experiment_output_folder is not None
            self.saving_folder = self.args.base_script_args.experiment_output_folder

        self.nn_trainer: NNPytorchTrainer = create_nn_trainer(
            args=self.args.nn_trainer_args,
            nn=self.nn_board_evaluator.net,
            saving_folder=self.saving_folder,
        )

        self.stockfish_boards_train = FenAndValueDataSet(
            file_name=self.args.dataset_args.train_file_name,
            preprocessing=self.args.dataset_args.preprocessing_data_set,
            transform_board_function=self.nn_board_evaluator.board_to_input_convert,
            transform_dataset_value_to_white_value_function=process_stockfish_value,  # pyright: ignore[reportUnknownArgumentType]
            transform_white_value_to_model_output_function=self.nn_board_evaluator.output_and_value_converter.from_value_white_to_model_output,
        )

        assert self.args.dataset_args.test_file_name is not None
        self.stockfish_boards_test = FenAndValueDataSet(
            file_name=self.args.dataset_args.test_file_name,
            preprocessing=self.args.dataset_args.preprocessing_data_set,
            transform_board_function=self.nn_board_evaluator.board_to_input_convert,
            transform_dataset_value_to_white_value_function=process_stockfish_value,  # pyright: ignore[reportUnknownArgumentType]
            transform_white_value_to_model_output_function=self.nn_board_evaluator.output_and_value_converter.from_value_white_to_model_output,
        )

        start_time = time.time()
        self.stockfish_boards_train.load()
        chipiron_logger.info("--- LOAD %s seconds --- ", time.time() - start_time)
        self.stockfish_boards_test.load()

        self.data_loader_stockfish_boards_train = DataLoader[FenAndValueData](
            self.stockfish_boards_train,
            batch_size=self.args.nn_trainer_args.batch_size_train,
            shuffle=True,
            num_workers=1,
            collate_fn=custom_collate_fn_fen_and_value,
        )

        self.data_loader_stockfish_boards_test = DataLoader(
            self.stockfish_boards_test,
            batch_size=self.args.nn_trainer_args.batch_size_test,
            shuffle=True,
            num_workers=1,
            collate_fn=custom_collate_fn_fen_and_value,
        )

        if self.args.base_script_args.testing:
            # mlflow.set_tracking_uri(
            #    uri=chipiron.utils.path_variables.ML_FLOW_URI_PATH_TEST
            # )
            with tempfile.TemporaryDirectory() as tmpdir:
                mlflow.set_tracking_uri(f"file:{tmpdir}")  # <- Key line!
        else:
            mlflow.set_tracking_uri(uri=chipiron.utils.path_variables.ML_FLOW_URI_PATH)

    def print_and_log_metrics(
        self, count_train_step: int, training_loss: float, test_error: float
    ) -> None:
        """
        Print and log training metrics to console and MLflow.

        Args:
            count_train_step (int): Current training step count.
            training_loss (float): Current training loss value.
            test_error (float): Current test error value.

        Returns:
            None
        """
        chipiron_logger.info(
            "count_train_step: %s, training loss: %s, lr: %s, time_elapsed: %s",
            count_train_step,
            training_loss,
            self.nn_trainer.scheduler.get_last_lr(),
            time.time() - self.start_time,
        )
        mlflow.log_metric(
            "training_loss",
            training_loss,
            step=count_train_step,
        )
        mlflow.log_metric(
            "test_error",
            test_error,
            step=count_train_step,
        )
        mlflow.log_metric(
            "lr",
            self.nn_trainer.scheduler.get_last_lr()[-1],
            step=count_train_step,
        )

    def run(self) -> None:
        """Running the learning of the NN.

        This method performs the training of the neural network. It iterates over the training data batches,
        computes the training loss, and updates the learning rate if necessary. It also prints the training
        loss and learning rate at regular intervals, and saves the learning process.

        Returns:
            None
        """
        chipiron_logger.info("Starting to learn the NN")

        with mlflow.start_run():
            params = asdict(self.args) | asdict(
                self.args.nn_trainer_args.neural_network_architecture_args
            )
            mlflow.log_params(params)

            # Log model summary.
            model_summary_file_name: str = os.path.join(
                self.saving_folder,
                "model_summary.txt",
            )
            with open(model_summary_file_name, "w", encoding="utf-8") as f:
                f.write(str(summary(self.nn_board_evaluator.net)))
            mlflow.log_artifact(model_summary_file_name)

            count_train_step = 0
            sum_loss_train: float = 0.0
            sum_loss_train_print: float = 0.0
            previous_dict: Any | None = None
            previous_train_loss: float | None = None

            fens_and_values_sample_batch: FenAndValueData
            for _ in range(self.args.nn_trainer_args.epochs_number):
                for __, fens_and_values_sample_batch in enumerate(
                    self.data_loader_stockfish_boards_train
                ):
                    # printing info to console
                    if count_train_step % 10000 == 0 and count_train_step > 0:
                        training_loss: float = sum_loss_train_print / 10000
                        sum_loss_train_print = 0
                        test_error: float = (
                            self.nn_trainer.compute_test_error_on_dataset(
                                data_test=self.data_loader_stockfish_boards_test
                            )
                        )
                        self.print_and_log_metrics(
                            count_train_step=count_train_step,
                            training_loss=training_loss,
                            test_error=test_error,
                        )

                    if (
                        count_train_step % 2000 == 0
                        and count_train_step > 0
                        and self.args.base_script_args.testing
                    ):
                        break

                    # every self.args['min_interval_lr_change'] steps we check for possibly decreasing the learning rate
                    if (
                        count_train_step
                        % self.args.nn_trainer_args.min_interval_lr_change
                        == 0
                        and count_train_step > 0
                    ):
                        # condition to decrease the learning rate
                        if (
                            previous_train_loss is not None
                            and sum_loss_train > previous_train_loss
                            and self.nn_trainer.scheduler.get_last_lr()[-1]
                            > self.args.nn_trainer_args.min_lr
                        ):
                            self.nn_trainer.scheduler.step()
                            chipiron_logger.info(
                                "decaying the learning rate to %s",
                                self.nn_trainer.scheduler.get_last_lr(),
                            )

                        chipiron_logger.info("count_train_step %s", count_train_step)
                        chipiron_logger.info(
                            "training loss: %s, sum_loss_train: %s",
                            sum_loss_train
                            / self.args.nn_trainer_args.min_interval_lr_change,
                            sum_loss_train,
                        )
                        chipiron_logger.info(
                            "previous_train_loss %s", previous_train_loss
                        )
                        chipiron_logger.info(
                            "learning rate %s", self.nn_trainer.scheduler.get_last_lr()
                        )

                        if previous_dict is not None:
                            diff_weighs = sum(
                                (x - y).abs().sum()
                                for x, y in zip(
                                    previous_dict.values(),
                                    self.nn_board_evaluator.net.state_dict().values(),
                                    strict=False,
                                )
                            )
                            print("diff_weights", diff_weighs)
                        previous_dict = copy.deepcopy(
                            self.nn_board_evaluator.net.state_dict()
                        )

                        previous_train_loss = sum_loss_train
                        sum_loss_train = 0.0

                    # MAIN: the training bit
                    count_train_step += 1
                    loss_train = self.nn_trainer.train(
                        fens_and_values_sample_batch.get_input_layer(),
                        fens_and_values_sample_batch.get_target_value(),
                    )
                    sum_loss_train += float(loss_train)
                    sum_loss_train_print += float(loss_train)

                    # saving the learning process
                    self.saving_things_to_file(
                        count_train_step=count_train_step,
                        X_train=fens_and_values_sample_batch.get_input_layer(),
                    )

    def saving_things_to_file(
        self, count_train_step: int, X_train: torch.Tensor
    ) -> None:
        """
        Saves the neural network parameters and trainer to file.

        Args:
            count_train_step (int): The current training step count.

        Returns:
            None
        """

        if count_train_step % self.args.nn_trainer_args.saving_interval == 0:
            safe_nn_param_save(
                nn=self.nn_board_evaluator.net,
                nn_param_folder_name=self.saving_folder,
                file_name=self.args.nn_trainer_args.neural_network_architecture_args.filename(),
            )
            safe_nn_architecture_save(
                nn_architecture_args=self.args.nn_trainer_args.neural_network_architecture_args,
                nn_param_folder_name=self.saving_folder,
            )
            safe_nn_trainer_save(self.nn_trainer, self.saving_folder)

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

            signature: ModelSignature = infer_signature(
                X_train.numpy(),
                self.nn_board_evaluator.net(X_train.to(device).detach())
                .cpu()
                .detach()
                .numpy(),
            )
            mlflow.pytorch.log_model(
                self.nn_board_evaluator.net,
                "model",
                signature=signature,
                conda_env=mlflow.pytorch.get_default_conda_env(),
            )

        if (
            self.args.nn_trainer_args.saving_intermediate_copy
            and count_train_step
            % self.args.nn_trainer_args.saving_intermediate_copy_interval
            == 0
        ):
            safe_nn_param_save(
                nn=self.nn_board_evaluator.net,
                nn_param_folder_name=self.saving_folder,
                training_copy=True,
            )

    def terminate(self) -> None:
        """
        Finishing the script. Profiling or timing.
        """
        self.base_script.terminate()

    @classmethod
    def get_args_dataclass_name(cls) -> type[LearnNNScriptArgs]:
        """
        Returns the dataclass type that holds the arguments for the script.

        Returns:
            type: The dataclass type for the script's arguments.
        """
        return LearnNNScriptArgs



"""
Module for the GamePlayer class.
"""

import copy
import queue

import chess

from chipiron.players.move_selector.move_selector import MoveRecommendation
from chipiron.utils import Seed
from chipiron.utils.communication.player_game_messages import MoveMessage
from chipiron.utils.dataclass import IsDataclass
from chipiron.utils.logger import chipiron_logger

from ..environments.chess_env.board.utils import FenPlusHistory
from .player import Player


class GamePlayer:
    """A class that wraps a player for a game purposes
    it adds the color information and probably stuff to continue the computation when the opponent is computing
    """

    _player: Player
    color: chess.Color

    def __init__(self, player: Player, color: chess.Color) -> None:
        self.color = color
        self._player = player

    @property
    def player(self) -> Player:
        """Return the player object.

        Returns:
            Player: The player object.
        """
        return self._player

    def select_move(
        self, fen_plus_history: FenPlusHistory, seed_int: Seed | None = None
    ) -> MoveRecommendation:
        """Selects the best move to play based on the current board position.

        Args:
            seed_int (seed | None, optional): The seed value for randomization. Defaults to None.

        Raises:
            Exception: If there are no legal moves in the current position.

        Returns:
            MoveRecommendation: The recommended move to play.
        """

        assert seed_int is not None
        assert self.color == fen_plus_history.current_turn()
        best_move: MoveRecommendation = self._player.select_move(
            fen_plus_history=fen_plus_history, seed_int=seed_int
        )
        return best_move


def game_player_computes_move_on_board_and_send_move_in_queue(
    fen_plus_history: FenPlusHistory,
    game_player: GamePlayer,
    queue_move: queue.Queue[IsDataclass],
    seed_int: Seed,
) -> None:
    """Computes the move for the game player on the given board and sends the move in the queue.

    Args:
        game_player (GamePlayer): The game player.
        queue_move (queue.Queue[IsDataclass]): The queue to send the move.
        seed_int (seed): The seed for move selection.

    Returns:
        None
    """

    if fen_plus_history.current_turn() == game_player.color:
        move_recommendation: MoveRecommendation = game_player.select_move(
            fen_plus_history=fen_plus_history, seed_int=seed_int
        )
        message: MoveMessage = MoveMessage(
            move=move_recommendation.move,
            corresponding_board=fen_plus_history.current_fen,
            player_name=game_player.player.id,
            evaluation=move_recommendation.evaluation,
            color_to_play=game_player.color,
        )
        chipiron_logger.info("Sending %s", message)

        deep_copy_message = copy.deepcopy(message)
        queue_move.put(deep_copy_message)
    else:
        chipiron_logger.warning(
            "Game player rejects move query %s",
            fen_plus_history.current_fen,
        )
